from weakref import WeakSet
import colorlog
from tqdm import tqdm
from .utils import classproperty
from .geo_default_types import create_geo_classes


logger = colorlog.getLogger('PySimultan')


class GeometryModel(object):

    _cls_instances = WeakSet()      # weak set with all created objects
    _create_all = False             # if true all properties are evaluated to create python objects when initialized

    @classproperty
    def _cls_instances_dict(cls):
        return dict(zip([x.id for x in cls._cls_instances], [x() for x in cls._cls_instances]))

    @classproperty
    def cls_instances(cls):
        return list(cls._cls_instances)

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        if "_cls_instances" not in cls.__dict__:
            cls._cls_instances = WeakSet()
        try:
            cls._cls_instances.add(instance)
        except Exception as e:
            logger.error(f'Error adding instance {instance} to _cls_instances: {e}')

        return instance

    def __init__(self, *args, **kwargs):
        self._wrapped_obj = kwargs.get('wrapped_obj', None)
        self.template_parser = kwargs.get('template_parser', None)

        self._vertices = kwargs.get('vertices', None)
        self._edges = kwargs.get('edges', None)
        self._edge_loops = kwargs.get('edge_loops', None)
        self._faces = kwargs.get('faces', None)
        self._volumes = kwargs.get('volumes', None)
        self._layers = kwargs.get('layers', None)

        self._geo_types = kwargs.get('geo_types', None)

        self.GeoBaseClass, self.LayerCls, self.VertexCls, self.EdgeCls, self.EdgeLoopCls, self.FaceCls, self.VolumeCls = create_geo_classes(self._geo_types)

        self.loaded = False

        # self.load_all()

    @property
    def id(self):
        return self._wrapped_obj.Id

    @property
    def filename(self):
        return self._wrapped_obj.File.Name

    @property
    def name(self):
        if self._wrapped_obj is not None:
            return self._wrapped_obj.Name

    @name.setter
    def name(self, value):
        if self._wrapped_obj is not None:
            self._wrapped_obj.Name = value

    @property
    def layers(self):
        if not self.loaded:
            self.load_all()
        if self._layers is None:
            self._layers = self.get_layers()
        return self._layers

    @property
    def vertices(self):
        if not self.loaded:
            self.load_all()

        if self._vertices is None:
            self._vertices = self.get_vertices()
        return self._vertices

    @property
    def edges(self):
        if not self.loaded:
            self.load_all()
        if self._edges is None:
            self._edges = self.get_edges()
        return self._edges

    @property
    def edge_loops(self):
        if not self.loaded:
            self.load_all()
        if self._edge_loops is None:
            self._edge_loops = self.get_edge_loops()
        return self._edge_loops

    @property
    def faces(self):
        if not self.loaded:
            self.load_all()
        if self._faces is None:
            self._faces = self.get_faces()
        return self._faces

    @property
    def volumes(self):
        if not self.loaded:
            self.load_all()
        if self._volumes is None:
            self._volumes = self.get_volumes()
        return self._volumes

    def __getattribute__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except KeyError:
            wrapped = object.__getattribute__(self, '_wrapped_obj')
            if wrapped is not None:
                return object.__getattribute__(wrapped, attr)
            else:
                raise KeyError

    def __setattr__(self, attr, value):

        if hasattr(self, '_wrapped_obj'):

            if hasattr(self._wrapped_obj, attr) and (self._wrapped_obj is not None):
                object.__setattr__(self._wrapped_obj, attr, value)
            else:
                object.__setattr__(self, attr, value)
        else:
            object.__setattr__(self, attr, value)

    def get_vertices(self):
        return [self.VertexCls(wrapped_obj=x, geometry_model=self) for x in tqdm(self._wrapped_obj.Geometry.Vertices.Items)]

    def get_edges(self):
        return [self.EdgeCls(wrapped_obj=x, geometry_model=self) for x in tqdm(self._wrapped_obj.Geometry.Edges.Items)]

    def get_edge_loops(self):
        return [self.EdgeLoopCls(wrapped_obj=x, geometry_model=self) for x in tqdm(self._wrapped_obj.Geometry.EdgeLoops.Items)]

    def get_faces(self):
        return [self.FaceCls(wrapped_obj=x, geometry_model=self) for x in tqdm(self._wrapped_obj.Geometry.Faces.Items)]

    def get_volumes(self):
        return [self.VolumeCls(wrapped_obj=x, geometry_model=self) for x in tqdm(self._wrapped_obj.Geometry.Volumes.Items)]

    def get_layers(self):
        return [self.LayerCls(wrapped_obj=x, geometry_model=self) for x in tqdm(self._wrapped_obj.Geometry.Layers.Items)]

    def get_face_by_id(self, id):

        face = self.FaceCls.get_obj_by_id(id)
        if face is None:
            _ = self.faces
        face = self.FaceCls.get_obj_by_id(id)
        return face

    def get_zone_by_id(self, id):

        zone = self.VolumeCls.get_obj_by_id(id)
        if zone is None:
            _ = self.volumes
        zone = self.VolumeCls.get_obj_by_id(id)
        return zone

    def load_all(self):

        self.loaded = True

        logger.info(f'Geometry model: {self.name}: loading vertices')
        _ = self.vertices
        logger.info(f'Geometry model: {self.name}: loading edges')
        _ = self.edges
        logger.info(f'Geometry model: {self.name}: loading edge loops')
        _ = self.edge_loops
        logger.info(f'Geometry model: {self.name}: loading faces')
        _ = self.faces
        logger.info(f'Geometry model: {self.name}: loading volumes')
        _ = self.volumes

        logger.info(f'\n\nGeometry model import info:\n----------------------------------------------')
        logger.info(f'Geometry model: {self.name}')
        logger.info(f'Number vertices: {self.vertices.__len__()}')
        logger.info(f'Number edges: {self.edges.__len__()}')
        logger.info(f'Number edge_loops: {self.edge_loops.__len__()}')
        logger.info(f'Number faces: {self.faces.__len__()}')
        logger.info(f'Number volumes: {self.volumes.__len__()}\n\n')

        self.loaded = True

    def get_geo_components(self, geo):
        """
        Get the simultan components linked to the geometric instance
        :param geo: geometry instance of type BaseGeoBaseClass
        :return: simultan components
        """
        return self.template_parser.get_geo_components(geo)

    def get_py_geo_components(self, geo, template_name=None):
        """

        :param geo: geometry instance of type BaseGeoBaseClass
        :param template_name: name of the template which should be used to create the component; if None suiting template is found automatically; default: None
        :return: python typed components
        """
        return self.template_parser.get_py_geo_components(geo, template_name=template_name)

    def create_layer(self, name, layer=None, **kwags):
        new_layer = self.LayerCls.create_new(self._wrapped_obj, name, layer=layer, **kwags)
        if self._layers is None:
            self._layers = []
        return new_layer

    def create_vertex(self, position, layer=None, **kwags):
        """
        Creates a new vertex in this geometry model
        :param position: np.array, tuple or list with three float; example: [0, 1, 12.25], np.array([0, 1, 12.25]), (0, 1, 12.25)
        :param layer: Instance of PySimultan.geo_default_types.BaseGeometricLayer
        :return: Instance of PySimultan.geo_default_types.BaseGeometricVertex
        """
        new_vertex = self.VertexCls.create_new(self._wrapped_obj, position, layer, **kwags)
        if self._vertices is None:
            self._vertices = []
        self._vertices.append(new_vertex)
        return new_vertex

    def create_edge(self, v1, v2, layer=None, **kwags):
        """
        Creates a new edge in this geometry model
        :param v1: Vertex 1; Instance of PySimultan.geo_default_types.BaseGeometricVertex
        :param v2: Vertex 2; Instance of PySimultan.geo_default_types.BaseGeometricVertex
        :param layer: Instance of PySimultan.geo_default_types.BaseGeometricLayer
        :return: Instance of PySimultan.geo_default_types.BaseGeometricEdge
        """
        new_edge = self.EdgeCls.create_new(self._wrapped_obj, v1, v2, layer, **kwags)
        if self._edges is None:
            self._edges = []

        self._edges.append(new_edge)
        return new_edge

    def create_edge_loop(self, edges, layer=None, **kwags):
        """
        Creates a new edge_loop in this geometry model
        :param edges: List, tuple or np.array with Instance of PySimultan.geo_default_types.BaseGeometricEdge
        :param layer: Instance of PySimultan.geo_default_types.BaseGeometricLayer
        :return: Instance of PySimultan.geo_default_types.BaseGeometriceEdgeLoop
        """
        new_edge_loop = self.EdgeLoopCls.create_new(self._wrapped_obj, edges, layer, **kwags)
        if self._edge_loops is None:
            self._edge_loops = []
        self._edge_loops.append(new_edge_loop)
        return new_edge_loop

    def create_face(self, edge_loop, orientation='forward', openings=None, layer=None, **kwags):
        """
        Creates a new face in this geometry model
        :param edge_loop: Instance of PySimultan.geo_default_types.BaseGeometriceEdgeLoop
        :param orientation: Orientation of the face; default is 'forward'; Valid values: 'forward', 'backward', 'undefined'
        GeometricOrientation.Backward, GeometricOrientation.Undefined
        :param openings: List, array or tuple with PySimultan.geo_default_types.BaseGeometriceEdgeLoop
        :param layer: Instance of PySimultan.geo_default_types.BaseGeometriceEdgeLoop
        :return: Instance of PySimultan.geo_default_types.BaseGeometriceFace
        """
        # if orientation == 'forward':
        #     orientation = GeometricOrientation.Forward
        # elif orientation == 'backward':
        #     orientation = GeometricOrientation.Backward
        # else:
        #     orientation = GeometricOrientation.Undefined

        new_face = self.FaceCls.create_new(self._wrapped_obj, edge_loop, orientation, openings, layer, **kwags)
        if self._faces is None:
            self._faces = []
        self._faces.append(new_face)
        return new_face

    def create_volume(self, faces, layer=None, **kwags):
        new_volume = self.VolumeCls.create_new(self._wrapped_obj, faces, layer, **kwags)
        if self._volumes is None:
            self._volumes = []
        self._volumes.append(new_volume)
        return new_volume

    def associate_face_with_component(self, component, face):
        """
        Associate a face with a component
        :param component: Instance of PySimultan.default_types.SimultanObject, PySimultan.default_types.List, PySimultan.default_types.ReferenceList
        :param face: Instance of PySimultan.geo_default_types.BaseGeometriceFace
        """
        face.associate_with_component(component, exch=self.template_parser.current_data_model.exch)

    def associate_volume_with_component(self, component, volume):
        """

        :param component:
        :param volume:
        """
        volume.associate_with_component(component, exch=self.template_parser.current_data_model.exch)



import sys
import time

import colorlog
import atexit
from uuid import uuid4
from functools import lru_cache
from weakref import WeakSet

from .SIMULTAN.Project.Services import IAuthenticationService
from .ParameterStructure import Projects
from .SIMULTAN.Project.Project import ExtendedProjectData
from .SIMULTAN.Project.ProjectLoaders import ZipProjectIO
from .SIMULTAN.UI.Services import ServicesProvider

# from .SIMULTAN.DataExchange import ComponentGeometryExchange
# from .ParameterStructure.Component import ComponentManagerType
from .ParameterStructure.Users import *
from .GeometryViewer import TemporaryGeometryViewerInstance
from .GeometryViewer.Service import GeometryViewerService
from .GeometryViewer.Model import *
from .GeometryViewer.IO import SimGeoIO
from .GeometryViewer import IGeometryViewerService
from System.Security import SecureString
from System.IO import FileInfo
from System import ArgumentOutOfRangeException
from .ParameterStructure.Components import SimComponent
from .ParameterStructure.Components import SimSlot, SimSlotBase, SimChildComponentEntry, ComponentUtils
from .ParameterStructure.Algorithms.Components import ComponentManagement
# from .ParameterStructure.Algorithms.Components import ComponentMapping
# from .ParameterStructure.Parameters import SimParameter
from .ParameterStructure.Values import SimMultiValueBigTable
# from .ParameterStructure.Assets import ContainedResourceFileEntry


ProjectData = Projects.ProjectData

# from .utils import *
# from .geo_default_types import *

# from .geometry import GeometryModel


logger = colorlog.getLogger('PySimultan')


class IAuthenticationServiceNew(IAuthenticationService):
    __namespace__ = "authenticate_namespace"

    user_name = None
    password = None

    def Authenticate(self, user_manager, project_file):
        # user_name = 'admin'
        # password = 'admin'

        sec_str = SecureString()
        for char in self.password:
            sec_str.AppendChar(char)

        user = user_manager.Authenticate(self.user_name, sec_str)

        user_manager.CurrentUser = user.Item1
        user_manager.EncryptionKey = user.Item2

        return user.Item1


data_models = WeakSet()


class DataModel:

    def __new__(cls, *args, **kwargs):

        instance = super().__new__(cls)

        try:
            data_models.add(instance)
        except Exception as e:
            logger.error(f'Error adding instance {instance} to data_models: {e}')
        return instance

    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """
        self.user_name = kwargs.get('user_name', 'admin')
        self.password = kwargs.get('password', 'admin')

        atexit.register(self.cleanup)

        self.id = uuid4()
        self.data = None
        self._project_data_manager = None
        self._user = None
        self._project = None
        self._zip_loader = None

        self.project_path = kwargs.get('project_path', None)

        self.service_provider = ServicesProvider()

        self.i_aut_service = IAuthenticationServiceNew
        self.i_aut_service.user_name = self.user_name
        self.i_aut_service.password = self.password

        # self.i_aut_service = create_IAuthenticationService(self.user_name, self.password)
        self.service_provider.AddService[IAuthenticationService](self.i_aut_service())

        self.serv = GeometryViewerService([], self.service_provider)

        self.service_provider.AddService[IGeometryViewerService](self.serv)

        self.exch = self.project.AllProjectDataManagers.GeometryCommunicator
        self.exch.ModelStore = self.serv
        self.inst = TemporaryGeometryViewerInstance(self.exch)
        self.resources = {}
        self.models_dict = {}

        if self.project_data_manager.AssetManager.Resources.__len__() > 0:
            for resource in self.project_data_manager.AssetManager.Resources:
                if resource is None:
                    continue
                self.resources[resource.Key] = resource
                self.models_dict[resource.Key] = None
                current_full_path = resource.CurrentFullPath
                if current_full_path == '?':
                    continue

                file_info = FileInfo(resource.CurrentFullPath)

                if file_info.Extension == '.simgeo':
                    model = SimGeoIO.Load(file_info, self.inst, self.serv)
                    self.models_dict[resource.Key] = model
                    try:
                        self.serv.AddGeometryModel(model)
                    except ArgumentOutOfRangeException as e:
                        logger.warning(f'Error while loading Model: {model} from {model.File}: {e}. Trying reload...')
                        model = SimGeoIO.Load(file_info, self.inst, self.serv)
                        self.models_dict[resource.Key] = model
                        self.serv.AddGeometryModel(model)

        self.ValueFields = self.project_data_manager.ValueManager.Items
        self.import_data_model()

    @property
    def models(self):
        return self.models_dict.values()

    @property
    def project_data_manager(self):
        if (self._project_data_manager) is None and (self.user is not None):
            self._project_data_manager = ExtendedProjectData()
        return self._project_data_manager

    @project_data_manager.setter
    def project_data_manager(self, value):
        self._project_data_manager = value

    @property
    def user(self):
        if self._user is None:
            self._user = SimUserRole.ADMINISTRATOR
        return self._user

    @user.setter
    def user(self, value):
        if value != self._user:
            self.project_data_manager = None
            self._project = None
        self._user = value

    @property
    def project(self):
        if (self._project is None) and (self.project_path is not None) and (self.project_data_manager is not None):
            logger.debug('loading project')
            self.project = ZipProjectIO.Load(FileInfo(self.project_path), self.project_data_manager, self.service_provider)
            exit_code = ZipProjectIO.AuthenticateUserAfterLoading(self.project, self.project_data_manager, bytearray('ThWmZq4t6w9z$C&F', 'ascii'))
            if not exit_code:
                logger.error('Could not open project. Wrong user or password! Exiting program...')
            ZipProjectIO.OpenAfterAuthentication(self.project, self.project_data_manager)
            logger.debug('project loaded successfull')
        return self._project

    @project.setter
    def project(self, value):
        self._project = value

    def get_typed_data(self, template_parser, create_all=False):

        template_parser._create_all = create_all
        template_parser.current_data_model = self

        for cls in template_parser.template_classes.values():
            if create_all:
                cls._create_all = True
            else:
                cls._create_all = False

        data = []

        for item in self.data.Items:
            logger.info(f'Creating python object for: {item.Name}')

            data.append(template_parser.create_python_object(item))

        logger.info('\n\nType info: \n----------------------------------')
        logger.info(f'created {data.__len__()} top level instances')
        for cls in set(template_parser.template_classes.values()):
            if hasattr(cls, 'cls_instances'):
                logger.info(f'created {cls.cls_instances.__len__()} instances of type {cls.__name__}')
        return data

    def import_data_model(self):
        self.data = self.project_data_manager.Components
        self.data.set_EnableAsyncUpdates(False)

        return self.data

    @lru_cache(maxsize=None)
    def get_geo_instance(self, file_id, type, id):
        geo_model = self.models[file_id]
        objects = getattr(geo_model.Geometry, type)

        return next((x for x in objects.Items if x.Id == id), None)

    def add_component(self, component: SimComponent):
        self.data.Add(component)

    def save(self):
        ZipProjectIO.Save(self.project, False)

    def cleanup(self):
        """
        Close and cleanup project
        """
        logger.info('closing project...')
        try:
            if self._project is not None:
                if self._project.IsOpened:
                    ZipProjectIO.Close(self._project, False, True)
                if self._project.IsLoaded:
                    ZipProjectIO.Unload(self._project)
        except Exception as e:
            pass

    # def create_new_component(self):
    #
    #     ref_comp = self.data.Items[0]
    #
    #     new_comp = SimComponent()
    #     new_comp.Name = 'Test'
    #
    #     new_param = SimParameter('test_param', 'Unicorn', 15.268)
    #     new_comp.Parameters.Add(new_param)
    #
    #     sub_new_comp = SimComponent()
    #     sub_new_comp.CurrentSlot = SimSlotBase(ComponentUtils.COMP_SLOT_AREAS)
    #     sub_new_comp.Name = 'SubTest'
    #
    #     entry = SimChildComponentEntry(SimSlot(SimSlotBase(ComponentUtils.COMP_SLOT_AREAS), '15'),
    #                                    sub_new_comp)
    #     new_comp.Components.Add(entry)
    #
    #     slot = SimSlot(ref_comp.CurrentSlot, '11')
    #     ComponentManagement.AddReferencedComponentSlot(new_comp, slot, self.user)
    #     ComponentManagement.AddReferencedComponent(new_comp, slot, ref_comp, self.user)
    #
    #     self.add_component(new_comp)
    #     self.save()
    #
    #     return new_comp

    def add_component_reference(self, comp: SimComponent, ref_comp: SimComponent, slot_extension: str, slot_name: str):

        slot = SimSlot(slot_name, str(slot_extension))
        ComponentManagement.AddReferencedComponentSlot(comp, slot, self.user)
        ComponentManagement.AddReferencedComponent(comp, slot, ref_comp, self.user)

    def add_new_geometry_model(self, file_name: str, model_name: str = None, return_resource=False):
        """
        Create and add a new geometry model
        :param file_name: name of the created .simgeo file
        :param model_name: name of the geometry model
        :param return_resource: return the resource
        :return: GeometryViewer.Model.GeometryModel, geo_resource
        """
        geo_resource = self.add_geometry_resource(file_name)
        file_info = FileInfo(geo_resource.CurrentFullPath)
        try:
            model = SimGeoIO.Load(file_info, self.inst, self.serv)
            self.models_dict[geo_resource.Key] = model
            self.serv.AddGeometryModel(model)
        except ArgumentOutOfRangeException as e:
            logger.warning(f'Error while loading Model: {model} from {model.File}: {e}. Trying reload...')
            model = SimGeoIO.Load(file_info, self.inst, self.serv)
            self.models_dict[geo_resource.Key] = model
            self.serv.AddGeometryModel(model)

        if model_name is not None:
            model.Name = model_name

        if return_resource:
            return model, geo_resource
        else:
            return model

    def add_geometry_resource(self, model_name: str):
        """
        Add / create new geometry resource (.simgeo file)
        :param model_name: name of the new .simgeo file without file extension; Example: 'new_model'
        """
        self.service_provider.GetService[IGeometryViewerService]()
        new_resource = self.project.AddEmptyGeometryResource(self.project.ProjectUnpackFolder, model_name, self.service_provider)

        return new_resource

    def add_resource(self, filename: str):
        return self.project.CopyResourceAsContainedFileEntry(FileInfo(filename), self.project.ProjectUnpackFolder)

    def add_table(self, table: SimMultiValueBigTable):
        self.project_data_manager.ValueManager.Add(table)

    def __del__(self):
        self.cleanup()


# if __name__ == '__main__':
#
#     # create example templates
#     templates = create_example_template_bim_bestand_network()
#
#     # write the example templates to a file:
#     with open('example_templates.yml', mode='w') as f_obj:
#         yaml.dump(templates, f_obj)
#
#     # load the example templates:
#     templates = load_templates('example_templates.yml')
#
#     # create classes from the templates:
#     template_classes = create_template_classes(templates)
#
#     simultan_components = create_example_simultan_components(templates, n=5)
#
#     simultan_components = class_type_simultan_components(simultan_components, template_classes)
#
#     # the simultan components are now of the type which is defined in the templates
#     print(simultan_components)
#
#     # the class typed components still keep all methods and attributes from simultan:
#     print(simultan_components[0].simultan_method())
#
#     # and the class typed components have the new defined method python_spec_func:
#     simultan_components[10].python_spec_func()



import typing
from typing import List as TypeList

from System.Windows.Media.Media3D import Point3D
from System import Array
from .GeometryViewer.Model.Geometry import Vertex, Edge, EdgeLoop, Face, Volume, Layer, PEdge, GeometricOrientation
from .GeometryViewer.Model import GeometryModel


GeometryModelType = typing.NewType('GeometryModel', GeometryModel)
EdgeModelType = typing.NewType('Edge', Edge)
EdgeLoopType = typing.NewType('EdgeLoop', EdgeLoop)
GeometricOrientationType = typing.NewType('GeometricOrientation', GeometricOrientation)


def create_layer(geometry_model: GeometryModelType, name: str):
    """
    Create a new layer in a geometry model
    :param geometry_model: GeometryViewer.Model.GeometryModel
    :param name: Name of the model; string
    :return: GeometryViewer.Model.Geometry.Layer
    """
    new_layer = Layer(geometry_model.Geometry, name)
    geometry_model.Geometry.Layers.Add(new_layer)
    return new_layer


def create_vertex(layer, position: typing.Iterable[float]):
    """
    Create a new vertex on a layer
    :param layer: GeometryViewer.Model.Geometry.Layer
    :param position: x, y, z position of the vertex, Iterable[float]
    :return: GeometryViewer.Model.Geometry.Vertex
    """
    return Vertex(layer, Point3D(position[0], position[1], position[2]))


def create_edge(layer, v1, v2):
    """
    Create a new edge on a layer
    :param layer: GeometryViewer.Model.Geometry.Layer
    :param v1: Vertex 1; GeometryViewer.Model.Geometry.Vertex
    :param v2: Vertex 2; GeometryViewer.Model.Geometry.Vertex
    :return: GeometryViewer.Model.Geometry.Edge
    """
    return Edge(layer, Array[Vertex]([v1, v2]))


def create_edge_loop(layer, edges: TypeList[EdgeModelType]):
    """
    Create a new edge loop on a layer
    :param layer: GeometryViewer.Model.Geometry.Layer
    :param edges: Iterable of GeometryViewer.Model.Geometry.Edges
    :return: GeometryViewer.Model.Geometry.EdgeLoop
    """
    return EdgeLoop(layer, Array[Edge](edges))


def create_face(layer, edge_loop, orientation: GeometricOrientationType, openings: TypeList[EdgeLoopType] = None):
    """
    Create a new face on a layer
    :param layer: GeometryViewer.Model.Geometry.Layer
    :param edge_loop: GeometryViewer.Model.Geometry.EdgeLoop
    :param orientation: GeometryViewer.Model.Geometry.GeometricOrientation
    :param openings: Iterable of GeometryViewer.Model.Geometry.EdgeLoop; optional
    :return: GeometryViewer.Model.Geometry.Face
    """
    if openings is None:
        openings = []

    return Face(layer, edge_loop, orientation, Array[EdgeLoop](openings))


def create_volume(layer, faces):
    """
    Create a new volume on a layer
    :param layer: GeometryViewer.Model.Geometry.Layer
    :param faces: Iterable of GeometryViewer.Model.Geometry.Face
    :return: Iterable of GeometryViewer.Model.Geometry.Volume
    """
    return Volume(layer, Array[Face](faces))



import numpy as np
import pandas as pd

from typing import List as TypeList
from typing import NewType
from typing import Union

from System.Collections.Generic import List as CSList
from System import Array, Double

from .ParameterStructure.Values import SimMultiValueBigTable, SimMultiValueBigTableHeader
from .ParameterStructure.Components import SimComponent
from .ParameterStructure.Components import SimSlot, SimSlotBase, SimChildComponentEntry, ComponentUtils
from .ParameterStructure.Algorithms.Components import ComponentMapping
from .ParameterStructure.Parameters import SimParameter
from .ParameterStructure.Values import SimMultiValueBigTable
from .ParameterStructure.Assets import ContainedResourceFileEntry


SimMultiValueBigTableType = NewType('SimMultiValueBigTable', SimMultiValueBigTable)
ContainedResourceFileEntryType = NewType('ContainedResourceFileEntry', ContainedResourceFileEntry)
SimParameterType = NewType('SimParameter', SimParameter)
SimComponentType = NewType('SimComponent', SimComponent)


def create_component(name=None, **kwargs) -> SimComponent:
    """
    Create a new Simultan component
    :param name: Name of the component; string
    :param kwargs: dictionary; set the components value of the key entry to to key value; Example: {'Visibility': 0, 'IsAutomaticallyGenerated': True}
    :return: ParameterStructure.Components.SimComponent
    """
    new_comp = SimComponent()

    if name is not None:
        new_comp.Name = name

    for key, value in kwargs.items():
        setattr(new_comp, key, value)

    return new_comp


def create_parameter(name='unnamed parameter', value: Union[int, float] = 0, unit:str='None'):
    """
    Creates a new SIMULTAN parameter

    :param name: Parameter name, string
    :param value: Value of the parameter; int or float
    :param unit: Parameter unit, string
    :return:
    """
    return SimParameter(name, unit, value)


def create_resource_reference(component: SimComponentType, resource: ContainedResourceFileEntryType):
    """
    Add a reference to a resource file to a component
    :param component: ParameterStructure.Components.SimComponent
    :param resource: ParameterStructure.Assets.ContainedResourceFileEntry
    :return:
    """
    return ComponentMapping.AddAsset(component, resource, '')


def assign_table_to_param(param: SimParameterType, table: SimMultiValueBigTable):
    """
    Assign a parameters value to a multi_value_pointer
    :param param: ParameterStructure.Components.SimParameter
    :param table: .ParameterStructure.Values.SimMultiValueBigTable
    """
    param.set_MultiValuePointer(table.DefaultPointer)


def add_sub_component(comp: SimComponentType, sub_comp: SimComponentType, slot_name: str, slot_extension: Union[str, int]):
    """
    Add a sub component to a component.

    :param comp: Component to add the subcomponent; ParameterStructure.Components.SimComponent
    :param sub_comp: Component to be added; ParameterStructure.Components.SimComponent
    :param slot_name: Slot name of the sub-component; string
    :param slot_extension: Slot extension of the sub-component; string
    """
    entry = SimChildComponentEntry(SimSlot(SimSlotBase(slot_name), str(slot_extension)),
                                   sub_comp)
    comp.Components.Add(entry)


def create_multi_value_big_table(table_name: str,
                                 column_title: str,
                                 column_headers: TypeList[str],
                                 column_units: TypeList[str],
                                 row_title: str,
                                 row_units: TypeList[str],
                                 row_headers: TypeList[str],
                                 values: np.ndarray) -> SimMultiValueBigTableType:
    """
    Create a multi value big table
    :param table_name: name of the
    :param column_title: column title
    :param column_headers: header of each column
    :param column_units: unit for each column
    :param row_title: row title
    :param row_headers: header of each row
    :param row_units: unit for each row
    :param values: table values
    :return: ParameterStructure.Values.SimMultiValueBigTable
    """

    column_headers_array = Array[SimMultiValueBigTableHeader](
        [SimMultiValueBigTableHeader(str(x[0]), str(x[1])) for x in zip(column_headers, column_units)])
    row_headers_array = Array[SimMultiValueBigTableHeader](
        [SimMultiValueBigTableHeader(str(x[0]), str(x[1])) for x in zip(row_headers, row_units)])

    def make_list(values):
        row = CSList[Double]()
        [row.Add(x) for x in values]
        return row

    rows = np.apply_along_axis(make_list, axis=1, arr=values)
    table_values = CSList[CSList[Double]]()
    [table_values.Add(x) for x in rows]

    table = SimMultiValueBigTable(table_name,
                                  column_title,
                                  row_title,
                                  column_headers_array,
                                  row_headers_array,
                                  table_values)

    return table


def df_to_multi_value_big_table(table_name: str,
                                df: pd.DataFrame,
                                column_title: str = '',
                                column_units: TypeList[str] = None,
                                row_title: str = '',
                                row_units: TypeList[str] = None):
    """
    Create a multi value big table from a pandas Dataframe
    :param df: dataframe
    :param table_name: name of the
    :param column_title: column title
    :param column_units: unit for each column
    :param row_title: row title
    :param row_units: unit for each row
    :return: ParameterStructure.Values.SimMultiValueBigTable
    """
    if column_units is None:
        column_units = [''] * df.columns.shape[0]

    if row_units is None:
        row_units = [''] * df.index.shape[0]

    table = create_multi_value_big_table(table_name=table_name,
                                         column_title=column_title,
                                         row_title=row_title,
                                         column_headers=df.columns,
                                         column_units=column_units,
                                         row_headers=df.index,
                                         row_units=row_units,
                                         values=df.values)

    return table



from .config import logger
from random import choices
# from .config import yaml
import re
from typing import Tuple


from System.Windows.Media import Color


class classproperty(object):

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class SimultanComponent(object):

    def __init__(self, *args, **kwargs):
        self.data = list(kwargs.get('data'))
        self.component_list = set()
        self.typed_component_list = set()

        self.get_component_list(self.data, self.data[0])
        self.get_typed_component_list()

    def get_component_list(self, components, component):
        if component not in self.typed_component_list:
            self.component_list.add(component)

            component_list = component.Components
            for neighbour in component_list:
                self.get_component_list(component_list, neighbour)

    def get_typed_component_list(self):
        for component in self.component_list:
            for param in component.Parameters.Items:
                if param.Name == 'TYPE':
                    self.typed_component_list.add(component)

    #
    # def dfs(self, visited, components, component):
    #     if component not in visited:
    #         print(component)
    #         visited.add(component)
    #         component_list = component.ContainedComponentsAsList
    #         for neighbour in component_list:
    #             dfs(visited, component_list, neighbour)
    #
    # def visit_components_list(self, visited, components, component, typed_component_list=[]):
    #     if component not in visited:
    #         visited.add(component)
    #
    #         for neighbour in components:
    #             if neighbour.ContainedParameters:
    #                 print('PARAMS')
    #                 print(neighbour.ContainedParameters)
    #                 for el in component.ContainedParameters:
    #                     if el.Name == 'TYPE':
    #                         print('TYPE')
    #                         print(el)
    #                         typed_component_list.append(neighbour)
    #             if neighbour.ContainedComponentsAsList:
    #                 print('LIST')
    #                 print(neighbour.ContainedComponentsAsList)
    #                 visit_components_list(component.ContainedComponentsAsList, neighbour, typed_component_list)
    #
    # def get_flat_list(self, visited, component_list, component):
    #     if component not in visited:
    #         visited.add(component)
    #         component_list = component.ContainedComponentsAsList
    #         for neighbour in component_list:
    #             dfs(visited, component_list, neighbour)
    #
    #     if not component_list:
    #         return []
    #     for comp in component_list:
    #         flat_list.append(comp)
    #         return (get_flat_list((comp.ContainedComponentsAsList, flat_list)))


def create_example_template():

    from .template_tools import Template

    material_template = Template(template_name='Material',
                                 template_id='1',
                                 content=['c', 'w20', 'w80', 'lambda', 'mu', 'rho'],
                                 documentation='c: specific heat = capacity in J/kg*K; ',
                                 units={'c': 'J/kg K',
                                        'w20': 'g/m=C2=B3',
                                        'w80': 'g/m=C2=B3',
                                        'lambda': 'W/mK',
                                        'mu': '-',
                                        'rho': 'kg/m=C2=B3'},
                                 types={'c': 'float',
                                        'w20': 'float',
                                        'w80': 'float',
                                        'lambda': 'float',
                                        'mu': 'float',
                                        'rho': 'float'}
                                 )

    layer_template = Template(template_name='Layer',
                              template_id='2',
                              content=['d', 'Material'],
                              documentation="d: thickness of the layer = in m, Material: see Template 'Material'",
                              units={'d': 'm', 'Material': '-'},
                              types={'d': 'float'}
                              )

    layer_template2 = Template(inherits_from=layer_template,
                               template_name='Layer2',
                               template_id='2.1',
                               content=['d', 'Material'],
                               documentation="d: thickness of the = layer in m, Material: see Template 'Material'",
                               units={'d': 'm', 'Material': '-'},
                               types={'d': 'int'}
                               )

    construction_template = Template(template_name='Construction',
                                     template_id='3',
                                     content=['layers'],
                                     documentation="layers: list of = items with type 'Layer'",
                                     units={'layers': '-'}
                                     )

    return [material_template, layer_template, layer_template2, construction_template]


def create_example_template_bim_bestand_network():

    from .template_tools import Template

    component_template = Template(template_name='Component',
                                  template_id=1,
                                  content=['AKS-Id', 'Loss Factor'],
                                  documentation='c: specific heat = capacity in J/kg*K;',
                                  units={'AKS-Id': 'J/kg K',
                                         'Loss Factor': 'g/m=C2=B3'},
                                  types={'AKS-Id': 'int',
                                         'Loss Factor': 'float'}
                                  )

    material_template = Template(template_name='Material',
                                 template_id='1',
                                 content=['c', 'w20', 'w80', 'lambda', 'mu', 'rho'],
                                 documentation='c: specific heat = capacity in J/kg*K; ',
                                 units={'c': 'J/kg K',
                                        'w20': 'g/m=C2=B3',
                                        'w80': 'g/m=C2=B3',
                                        'lambda': 'W/mK',
                                        'mu': '-',
                                        'rho': 'kg/m=C2=B3'},
                                 types={'c': 'float',
                                        'w20': 'float',
                                        'w80': 'float',
                                        'lambda': 'float',
                                        'mu': 'float',
                                        'rho': 'float'}
                                 )

    edge_template = Template(inherits_from=component_template,
                             template_name='Edge',
                             template_id='2',
                             content=['Start-ID', 'End-ID', 'COMPONENT-ID', 'COMPONENT-TYPE', 'Length', 'Lambda', 'K'],
                             documentation="d: thickness of the layer = in m, Material: see Template 'Material'",
                             units={'d': 'm', 'Material': '-'},
                             types={'d': 'float'}
                             )

    confuser_diffuser_rectangular_template = Template(inherits_from=edge_template,
                                                      template_name='CONFUSERDIFFUSERRECTANGULAR',
                                                      template_id='2.1',
                                                      content=['EndHight', 'EndWidth', 'StartHight', 'StartWidth'],
                                                      documentation="d: thickness of the = layer in m, Material: see Template 'Material'",
                                                      units={'d': 'm', 'Material': '-'},
                                                      types={'d': 'int'}
                                                      )

    confuser_diffuser_round_template = Template(inherits_from=edge_template,
                                                template_name='CONFUSERDIFFUSERROUND',
                                                template_id='3',
                                                content=['EndDiameter', 'StartDiameter'],
                                                documentation="layers: list of = items with type 'Layer'",
                                                units={'layers': '-'}
                                                )

    return [component_template, material_template, edge_template, confuser_diffuser_rectangular_template, confuser_diffuser_round_template]


def class_type_simultan_components(components, template_classes):

    # collect component classes in a dictionary:
    component_classes = {}

    # loop trough all components:
    for component in components:
        # get the template-id
        template_name = None
        for comp_type in component.Parameters.Items:
            if comp_type.Name == 'TYPE':
                template_name = comp_type.TextValue

        # check if the component class already exists:
        if template_name in component_classes.keys():     # if it already = exists take it
            new_component_class_dict = component_classes[template_name]
        elif template_name in template_classes.keys():   # create new component class
            # find the python template class:
            template_class = template_classes[template_name]

            # init new instance
            new_instance = template_class(wrapped_obj=component)

            print(new_instance)


    return components


def create_example_simultan_components(templates, flat_list_components):

    simultan_components = []

    for template in templates:
        for component in flat_list_components:
            print(template)
            print(component)

    return simultan_components


def get_obj_by_id(cls, id):
    return cls._cls_instances_dict.get(id, None)


def sort_component_list(components):

    if components.__len__() == 0:
        return components

    slots = [x.Slot.SlotBase.Base for x in components]
    if not all(slots[0] == x for x in slots):
        raise TypeError(f'List elements do not have same slot')

    # slot_extensions = [int(re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", x.CurrentSlot)[-1]) for x in components]
    slot_extensions = [int(x.Slot.SlotExtension) for x in components]
    return slot_extensions


def sort_references(references):

    if references.__len__() == 0:
        return references
    # check if same slot:
    slots = [x.ReferenceFunction.SlotBase for x in references]

    if not all(slots[0] == x for x in slots):
        raise TypeError(f'List elements do not have same slot')

    slot_extensions = [int(x.ReferenceFunction.SlotExtension) for x in references]
    return slot_extensions


def get_random_color():
    rgb = choices(range(256), k=3)
    return Color.FromArgb(rgb[0], rgb[1], rgb[2], 255)


def parse_slot(full_slot_name: str) -> Tuple[str, any]:
    """
    Parse the slot name. Returns slot_name and slot_extension
    :param full_slot_name:
    :return: (slot_name, slot_extension)
    """
    slot_extension = full_slot_name.split()[-1]
    slot_name = ' '.join(full_slot_name.split()[0:-1])

    return slot_name, slot_extension



from __future__ import annotations
from collections import UserList
from functools import lru_cache
import pandas as pd
import numpy as np
from weakref import WeakSet
import colorlog
import re

# from typing import List as TypeList
from typing import NewType, Union, Optional

# from System.Collections.Generic import List as CSList
# from System import Array, Double

from .ParameterStructure.Values import SimMultiValueBigTable, SimMultiValueBigTableHeader
from .ParameterStructure.Instances import SimInstancePlacementGeometry, SimInstanceType
from .ParameterStructure.Parameters import SimParameter
from .ParameterStructure.Components import SimComponent, SimSlot, SimSlotBase, SimChildComponentEntry, ComponentUtils
from .ParameterStructure.Algorithms.Components import ComponentManagement, ComponentMapping
from .ParameterStructure.Assets import ContainedResourceFileEntry
from .ParameterStructure.SimObjects import SimId

from .config import continue_on_error
from .utils import sort_references, sort_component_list, parse_slot
from .simultan_utils import create_parameter
from . import create_component
from .slots import SimSlotBase
from .data_model import data_models


SimMultiValueBigTableType = NewType('SimMultiValueBigTable', SimMultiValueBigTable)
ContainedResourceFileEntryType = NewType('ContainedResourceFileEntry', ContainedResourceFileEntry)
SimParameterType = NewType('SimParameter', SimParameter)
SimComponentType = NewType('SimComponent', SimComponent)


logger = colorlog.getLogger('PySimultan')


class classproperty(object):

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def init_parameter(cls, comp, param_name, val, user, slot_parameter_variable=False):

    if val is None:
        return
    if isinstance(val, (int, float, str)):
        if param_name in cls._slots.keys():
            if not slot_parameter_variable:
                raise TypeError((f'Error initializing instance of {cls.__name__}:\n'
                                 f'{param_name} is associated with a slot.'
                                 f'Valid parameter values are SimComponent or SimultanObject.'
                                 f'but the value is of type {type(val)}.\n'
                                 f'If this is intentional, initialize with option slot_parameter_variable=True'))

        if hasattr(comp, param_name):
            setattr(comp, param_name, val)
        else:
            # create parameter:
            unit = str(cls._units.get(param_name, ''))
            if isinstance(val, str):
                param = create_parameter(name=param_name,
                                         value=0.0,
                                         unit=unit)
                param.set_TextValue(val)
            else:
                param = create_parameter(name=param_name,
                                         value=float(val),
                                         unit=unit)
                param.set_ValueCurrent(float(val))
            comp.Parameters.Add(param)
    elif isinstance(val, (SimComponent, SimultanObject)):
        # if the parameter is associated with a slot:
        if param_name in cls._slots.keys():
            # create sub component or reference
            full_slot_name = cls._slots[param_name]
            slot_name, slot_extension = parse_slot(full_slot_name)

            if isinstance(val, SimultanObject) or (val is None):
                slot = SimSlot(SimSlotBase(slot_name), str(slot_extension))
                ComponentManagement.AddReferencedComponentSlot(comp, slot, user)
                if val is not None:
                    ComponentManagement.AddReferencedComponent(comp, slot, val._wrapped_obj, user)
            elif isinstance(val, SimComponent):
                entry = SimChildComponentEntry(SimSlot(SimSlotBase(slot_name), str(slot_extension)), val)
                comp.Components.Add(entry)
            else:
                raise TypeError(
                    f'Error setting parameter {param_name}. Wrong type {type(val)}. This parameter is associated with a slot. The parameter value must be of type {SimultanObject} or {SimComponent}')
        else:
            raise TypeError(
                f'Error setting parameter {param_name}. Can not create sub-component or reference. There is no slot defined for this parameter')
    else:
        raise TypeError(
            f'Error setting parameter {param_name}. Wrong type {type(val)}. The value must be a instance of int, float, str, SimultanObject, SimultanObject')


def create_simultan_component_from_template(cls, **kwargs):
    """

    :param cls:
    :param kwargs:
    @keyword create_undefined_parameters: if True create parameters for also for undefined keys
    @keyword slot_parameter_variable: if True automatically choose to create parameters, sub-components or references
    depending on the content value although a slot is defined in the template for the content
    :return:
    """

    create_undefined_parameters = kwargs.get('create_undefined_parameters', False)
    slot_parameter_variable = kwargs.get('slot_parameter_variable', False)

    comp = create_component()

    # add TYPE parameter:
    type_param = SimParameter('TYPE', '', 0.0)
    type_param.set_TextValue(cls._template_name)
    comp.Parameters.Add(type_param)

    user = kwargs['data_model'].user

    # create parameters:
    for param_name in cls._content:

        val = kwargs.get(param_name, None)
        init_parameter(cls, comp, param_name, val, user, slot_parameter_variable)

        # if val is None:
        #     continue
        # if isinstance(val, (int, float, str)):
        #     if param_name in cls._slots.keys():
        #         if not slot_parameter_variable:
        #             raise TypeError((f'Error initializing instance of {cls.__name__}:\n'
        #                              f'{param_name} is associated with a slot.'
        #                              f'Valid parameter values are SimComponent or SimultanObject.'
        #                              f'but the value is of type {type(val)}.\n'
        #                              f'If this is intentional, initialize with option slot_parameter_variable=True'))
        #
        #     if hasattr(comp, param_name):
        #         setattr(comp, param_name, kwargs.get(param_name))
        #     else:
        #         # create parameter:
        #         unit = str(cls._units.get(param_name, ''))
        #         if isinstance(val, str):
        #             param = create_parameter(name=param_name,
        #                                      value=0.0,
        #                                      unit=unit)
        #             param.set_TextValue(val)
        #         else:
        #             param = create_parameter(name=param_name,
        #                                      value=float(val),
        #                                      unit=unit)
        #             param.set_ValueCurrent(float(val))
        #         comp.Parameters.Add(param)
        # elif isinstance(val, (SimComponent, SimultanObject)):
        #     # if the parameter is associated with a slot:
        #     if param_name in cls._slots.keys():
        #         # create sub component or reference
        #         full_slot_name = cls._slots[param_name]
        #         slot_name, slot_extension = parse_slot(full_slot_name)
        #
        #         if isinstance(val, SimultanObject) or (val is None):
        #             slot = SimSlot(SimSlotBase(slot_name), str(slot_extension))
        #             user = kwargs['data_model'].user
        #             ComponentManagement.AddReferencedComponentSlot(comp, slot, user)
        #             if val is not None:
        #                 ComponentManagement.AddReferencedComponent(comp, slot, val._wrapped_obj, user)
        #         elif isinstance(val, SimComponent):
        #             entry = SimChildComponentEntry(SimSlot(SimSlotBase(slot_name), str(slot_extension)), val)
        #             comp.Components.Add(entry)
        #         else:
        #             raise TypeError(f'Error setting parameter {param_name}. Wrong type {type(val)}. This parameter is associated with a slot. The parameter value must be of type {SimultanObject} or {SimComponent}')
        #     else:
        #         raise TypeError(
        #             f'Error setting parameter {param_name}. Can not create sub-component or reference. There is no slot defined for this parameter')
        # else:
        #     raise TypeError(f'Error setting parameter {param_name}. Wrong type {type(val)}. The value must be a instance of int, float, str, SimultanObject, SimultanObject')

    if create_undefined_parameters:
        logger.warning(f'Using option slot_parameter_variable. This is an experimental feature and may cause errors')
        undef_params = set(kwargs.keys()) - set(cls._content) - {'name',
                                                                 'current_slot',
                                                                 'create_undefined_parameters',
                                                                 'data_model'}

        for param_name in undef_params:
            val = kwargs.get(param_name, None)
            init_parameter(cls, comp, param_name, val, user, slot_parameter_variable)

    return comp


class SlotError(Exception):
    """Exception raised for errors concerning slot."""


class MetaMock(type):
    def __call__(cls, *args, **kwargs):
        """
        Metaclass to implement object initialization either with wrapped_obj or keywords.

        If a wrapped_obj is defined, create new SimultanObject which wraps a SIMULTAN component (wrapped_obj).

        If no 'wrapped_obj' is defined, a new SimComponent is created with the content defined in the template and the
        values are set to the values defined in kwargs.
        """

        obj = cls.__new__(cls, *args, **kwargs)

        wrapped_obj = kwargs.get('wrapped_obj', None)
        if wrapped_obj is None:

            data_model = kwargs.get('data_model', None)
            if data_model is None:
                if list(data_models.data).__len__() == 1:
                    data_model = list(data_models.data)[0]()
                    kwargs['data_model'] = data_model
                else:
                    raise TypeError((f'Error creating new instance of class {cls.__name__}:\n'
                                     f'Any data model was defined. Tried to use default data model but there are multiple datamodels.\n'
                                     f'Define the data model to use with the key: data_model'))

            wrapped_obj = create_simultan_component_from_template(cls, **kwargs)
            if 'name' in kwargs.keys():
                wrapped_obj.Name = kwargs.get('name')
            if 'current_slot' in kwargs.keys():
                value = kwargs.get('current_slot')
                if isinstance(value, str):
                    value = SimSlotBase(value)
                elif isinstance(value, SimSlotBase):
                    pass
                else:
                    raise TypeError('Current slot must be of type str or SimSlotBase')
            elif cls._default_slot is not None:
                value = SimSlotBase(cls._default_slot)
            wrapped_obj.set_CurrentSlot(value)

            data_model.add_component(wrapped_obj)

            init_dict = {}
            init_dict['data_model_id'] = data_model.id
            init_dict['template_parser'] = cls._template_parser
            init_dict['wrapped_obj'] = wrapped_obj
            obj.__init__(*args, **init_dict)
            # instance = cls(**init_dict)
        else:
            obj.__init__(*args, **kwargs)
        return obj


class SimultanObject(object, metaclass=MetaMock):

    # __metaclass__ = MetaMock

    _cls_instances = WeakSet()  # weak set with all created objects
    _create_all = False     # if true all properties are evaluated to create python objects when initialized
    _cls_instances_dict_cache = None

    @classproperty
    def _cls_instances_dict(cls):
        if cls._cls_instances_dict_cache is None:
            cls._cls_instances_dict_cache = dict(zip([x.id for x in cls._cls_instances], [x for x in cls._cls_instances]))
        return cls._cls_instances_dict_cache

    @classproperty
    def cls_instances(cls):
        return list(cls._cls_instances)

    def __new__(cls, *args, **kwargs):

        instance = super().__new__(cls)

        if "_cls_instances" not in cls.__dict__:
            cls._cls_instances = WeakSet()
        try:
            cls._cls_instances.add(instance)
            cls._cls_instances_dict_cache = None
        except Exception as e:
            logger.error(f'Error adding instance {instance} to _cls_instances: {e}')

        return instance

    def __init__(self, *args, **kwargs):
        """
        Initialize a SIMULTAN component from a template class

        There are two options to initialize a SimultanObject:

        1. with a 'wrapped_obj', 'template_parser' and 'data_model_id':
            if the keyword 'wrapped_obj' with a SimComponent is passed, a new SimultanObject is created, which wraps the
            passed SimComponent. Additionally the 'template_parser' and 'data_model_id' must be defined. This is mostly
            intended to be done by the template parser to create python objects.

            Example:
            new_instance = template_class(wrapped_obj=sim_component,
                                          template_parser=self,
                                          data_model_id=self._current_data_model.id)

        2. without a 'wrapped_obj' to create a new SimComponent:

            if no 'data_model' is passed, the default data model is used, which is the first loaded data model

            new_slot_comp_3 = SimultanObject(name='test_name',
                                             current_slot='Element',
                                             Param0='Test Text Value',
                                             Param1='Test Value',
                                             Param2=new_comp_1,
                                             Param3=new_comp_2,
                                             data_model=data_model)

        :param args:
        :param kwargs:

        To creat a new instance wrapping an existing SimComponent:
        @keyword wrapped_obj: SimComponent to be wrapped
        @keyword template_parser: Template parser instance
        @keyword data_model_id: Id of the data model the component belongs to

        If a new SimComponent is created:
        @keyword name: Name of the Component
        @keyword current_slot: Name of the Component's slot (for valid slots see SIMULTAN slot documentation)
        @keyword data_model: Id of the data model the component belongs to

        @keyword create_undefined_parameters: if True create parameters for also for undefined keys
        @keyword slot_parameter_variable: if True automatically choose to create parameters, sub-components or references
        """

        self._wrapped_obj = kwargs.get('wrapped_obj', None)
        self._contained_components = kwargs.get('contained_components', None)
        self._contained_parameters = kwargs.get('contained_parameters', None)
        self._flat_sub_comp_list = kwargs.get('flat_sub_comp_list', None)
        self._referenced_components = kwargs.get('referenced_components', None)
        self._template_parser = kwargs.get('template_parser', None)
        self._data_model_id = kwargs.get('data_model_id', None)

        self._slot_components = None

        if self._create_all:
            _ = self.contained_components
            _ = self.contained_parameters
            _ = self.referenced_components

        # _ = self.slot_components

    def __getattribute__(self, attr):
        try:
            return object.__getattribute__(self, attr)
        except (KeyError, AttributeError):
            wrapped = object.__getattribute__(self, '_wrapped_obj')
            if wrapped is not None:
                return object.__getattribute__(wrapped, attr)
            else:
                raise KeyError

    def __setattr__(self, attr, value):
        if hasattr(self, '_wrapped_obj'):

            if hasattr(self._wrapped_obj, attr) and (self._wrapped_obj is not None):
                object.__setattr__(self._wrapped_obj, attr, value)
            else:
                object.__setattr__(self, attr, value)
        else:
            object.__setattr__(self, attr, value)

        # if (attr in self.__dict__) or (attr in ['_wrapped_obj',
        #                                         '_contained_components',
        #                                         '_contained_parameters',
        #                                         '_flat_sub_comp_list',
        #                                         '_referenced_components',
        #                                         '_template_parser',
        #                                         '_data_model_id']):
        #
        #     object.__setattr__(self, attr, value)
        # else:
        #     object.__setattr__(self._wrapped_obj, attr, value)

    def __getstate__(self):

        logger.warning(f'__getstate__ in experimental state')

        obj_dict = dict()
        for key in dir(self.__class__):
            if type(getattr(self.__class__, key)) is property:
                print(key)
                obj_dict[key] = getattr(self, key)

        return obj_dict

    def __setstate__(self, state):

        logger.warning(f'__setstate__ in experimental state')

        for key, value in state.items():
            try:
                setattr(self, key, value)
            except AttributeError as e:
                pass

    def __repr__(self):
        return f'{self.name}: ' + object.__repr__(self)

    def __del__(self):
        self.__class__._cls_instances_dict_cache = None

    @property
    def id(self):
        if self._wrapped_obj is not None:
            return self._wrapped_obj.Id

    @property
    def name(self):
        if self._wrapped_obj is not None:
            return self._wrapped_obj.Name

    @name.setter
    def name(self, value: str):
        if self._wrapped_obj is not None:
            self._wrapped_obj.Name = value

    @property
    def contained_components(self):
        if self._contained_components is None:
            if self._wrapped_obj is not None:
                self._contained_components = [self._template_parser.create_python_object(x.Component) for x in self._wrapped_obj.Components.Items]
        return self._contained_components

    @property
    def contained_parameters(self):
        if self._contained_parameters is None:
            if self._wrapped_obj is not None:
                self._contained_parameters = {x.Name: x.get_ValueCurrent() for x in self._wrapped_obj.Parameters.Items}
        return self._contained_parameters

    # @property
    # def flat_sub_comp_list(self):
    #     if self._flat_sub_comp_list is None:
    #         if self._wrapped_obj is not None:
    #             self._flat_sub_comp_list = [self._template_parser.create_python_object(x) for x in self._wrapped_obj.GetFlatSubCompList()]
    #     return self._flat_sub_comp_list

    @property
    def referenced_components(self):
        if self._referenced_components is None:
            if self._wrapped_obj is not None:
                self._referenced_components = [self._template_parser.create_python_object(x.Reference) for x in self._wrapped_obj.ReferencedComponents.Items]
        return self._referenced_components

    @property
    def referenced_assets(self):
        return [x.Resource for x in self._wrapped_obj.ReferencedAssets.Items.Items]

    @lru_cache()
    def get_param_index(self, param: str) -> int:
        """
        Return the index of the parameter with the name 'param' in self self._wrapped_obj.Parameters.Items
        :param param: str; Example: 'Area'
        :return: Index of the parameter
        """
        idx = next((i for i, x in enumerate(self._wrapped_obj.Parameters.Items) if x.Name == param), None)
        return idx

    def get_param(self, param: str) -> SimParameterType:
        """
        Return the parameter of the parameter with the name 'param' in self self._wrapped_obj.Parameters.Items
        :param param: str; Example: 'Area'
        :return: SimParameter
        """
        idx = self.get_param_index(param)
        if idx is None:
            raise AttributeError(f'{self} {self.name} {self.id} has no parameter {param}')
        return self._wrapped_obj.Parameters.Items[idx].get_ValueCurrent()

    def set_param(self, param: str, value: float):
        """
        Set the value of a param
        :param param: name of the parameter
        :param value: value of the parameter
        """
        idx = self.get_param_index(param)
        if idx is None:
            raise AttributeError(f'{self} {self.name} has no parameter {param}')
        self._wrapped_obj.Parameters.Items[idx].set_ValueCurrent(value)

    @property
    def current_slot(self):
        return self._wrapped_obj.CurrentSlot

    @current_slot.setter
    def current_slot(self, value):
        if isinstance(value, str):
            value = SimSlotBase(value)
        elif isinstance(value, SimSlotBase):
            pass
        else:
            raise TypeError('Current slot must be of type str or SimSlotBase')
        self._wrapped_obj.set_CurrentSlot(value)

    @property
    def fits_in_slot(self):
        return self._wrapped_obj.CurrentSlot

    @property
    def slot_extension(self):
        """
        Returns the slot extension of the component
        :param
        """
        try:
            return int(re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", self.current_slot)[-1])
        except Exception as e:
            return None

    @property
    def slot_components(self):
        if self._slot_components is None:
            self._slot_components = self.get_slot_components()
        return self._slot_components

    @slot_components.setter
    def slot_components(self, value):
        self._slot_components = value

    def get_slot_components(self):
        """
        Return a DataFrame with the slot components.
        Dataframe columns: 'comp': component, 'slot': slot name, 'slot_extension': slot extension, 'comp_type': 0: subcomponent, 1: referenced component
        :return:
        """
        sc_df = pd.DataFrame(columns=['comp', 'slot', 'slot_extension', 'comp_type'])

        # subcomponents
        for comp in self._wrapped_obj.Components.Items:
            c_slot = comp.Slot.SlotBase.Base
            c_slot_extension = comp.Slot.SlotExtension

            sc_df = sc_df.append(
                pd.DataFrame(data={'comp': [self._template_parser.create_python_object(comp.Component)],
                                   'slot': [c_slot],
                                   'slot_extension': c_slot_extension,
                                   'comp_type': 0}),
                ignore_index=True)

        # referenced components
        for comp in self._wrapped_obj.ReferencedComponents.Items:
            c_slot = comp.ReferenceFunction.SlotBase.Base
            c_slot_extension = comp.ReferenceFunction.SlotExtension

            sc_df = sc_df.append(
                pd.DataFrame(data={'comp': [self._template_parser.create_python_object(comp.Reference)],
                                   'slot': [c_slot],
                                   'slot_extension': c_slot_extension,
                                   'comp_type': 1}),
                ignore_index=True)

        return sc_df

    def add_sub_component(self,
                          component: Union[SimComponentType, SimultanObject],
                          slot_name: str,
                          slot_extension: Union[int, float, str],
                          alter_slot=False):
        """
        Add a sub-component to this component
        :param component: Component to add; Type: SimComponent or SimultanObject
        :param slot_name: name of the slot of the created sub-component
        :param slot_extension: slot extension
        :param alter_slot: if True change the slot of the component if it does not match the current slot of the component
        """

        sub_comp_slot = SimSlot(SimSlotBase(slot_name), str(slot_extension))

        if isinstance(component, SimultanObject):
            component = component._wrapped_obj

        if component.CurrentSlot is not None:
            if alter_slot:
                component.CurrentSlot = SimSlotBase(slot_name)
            else:
                raise SlotError(f'Component: {component.Name, component.Id.GlobalId} has already asigned slot:\n{component.CurrentSlot.Base}\nThis slot does not match template\'s slot:\n{slot_name}')
        else:
            component.CurrentSlot = SimSlotBase(slot_name)

        if not component.Id == SimId.Empty:
            component.Id = SimId.Empty

        entry = SimChildComponentEntry(sub_comp_slot, component)
        self._wrapped_obj.Components.Add(entry)

        self.slot_components = self.get_slot_components()

    def remove_sub_component(self,
                             component: Union[SimComponent, SimultanObject] = None,
                             slot_name: Optional[str] = None,
                             slot_extension: Optional[Union[int, float, str]] = None):
        """
        Removes a sub component; Sub component defined either by component or slot_name and slot_extension
        :param component: Sub component to remove
        :param slot_name: [Optional]: slot name
        :param slot_extension: Optional]: slot extension
        """
        entry = None

        if isinstance(component, SimultanObject):
            component = component._wrapped_obj

        if component is not None:
            entry = next((x for x in self._wrapped_obj.Components.Items if x.Component == component), None)

        if None not in [slot_name, slot_extension, entry]:
            entry = next((x for x in self._wrapped_obj.Components.Items
                          if all([x.Slot.SlotBase.Base == slot_name,
                                  x.Slot.SlotExtension == slot_extension])),
                         None)

        if entry is not None:
            self._wrapped_obj.Components.Remove(entry)
        else:
            logger.warning(f'Could not remove sub component. Entry not found')
            raise ValueError(f'Could not remove sub component. Entry not found')

        self.slot_components = self.get_slot_components()

    def add_component_reference(self,
                                component: Union[SimComponentType, SimultanObject],
                                slot_name: str,
                                slot_extension: Union[int, float, str]):
        """
        Add a reference of component to self
        :param component: Instance of PySimultan.default_types.SimultanObject, PySimultan.default_types.List, PySimultan.default_types.ReferenceList
        :param slot_name: Name of the slot
        :param slot_extension:
        """
        slot = SimSlot(SimSlotBase(slot_name), str(slot_extension))
        user = self._template_parser.current_data_model.user
        ComponentManagement.AddReferencedComponentSlot(self._wrapped_obj, slot, user)
        if isinstance(component, SimultanObject):
            component = component._wrapped_obj
        ComponentManagement.AddReferencedComponent(self._wrapped_obj, slot, component, user)

        self.slot_components = self.get_slot_components()

    def remove_component_reference(self, component: Union[SimComponentType, SimultanObject]):
        """
        Removes the reference to a component.
        :param component: Component which's reference to remove
        """
        if isinstance(component, SimultanObject):
            component = component._wrapped_obj
        ComponentManagement.RemoveReferencedComponent_Level0(self._wrapped_obj, component, True)

        self.slot_components = self.get_slot_components()

    def add_parameter(self, name: str, unit: str, value: Union[int, float], param_type: str = 'int'):
        """
        Add a parameter to this component. Additionally a property is added to access the parameter from this component.
        :param name: name of the parameter
        :param unit: unit of the parameter; Example: 'kg'
        :param value: value of the parameter: Example: 15.875
        :param param_type: type of the parameter; if param_type == 'str': param.get_TextValue()
               else: param.get_ValueCurrent() is returned
        """
        from .template_tools import add_properties

        new_param = SimParameter(name, unit, float(value))
        self._wrapped_obj.Parameters.Add(new_param)
        object.__setattr__(self, name, add_properties(name, param_type=param_type))

    def add_asset(self, resource: ContainedResourceFileEntryType):
        """
        Add a reference to a resource (file) to this component
        :param resource: Instance of ParameterStructure.Assets.ContainedResourceFileEntry
        """
        ComponentMapping.AddAsset(self._wrapped_obj, resource, '')

    def assign_table_to_parameter(self, param_name: str, table: SimMultiValueBigTableType):
        """
        set a prarameter's value to a SimMultiValueBigTable value
        :param param_name: name of the parameter
        :param table: ParameterStructure.Values.SimMultiValueBigTable
        """
        idx = self.get_param_index(param_name)
        param = self._wrapped_obj.Parameters.Items[idx]
        param.set_MultiValuePointer(table.DefaultPointer)

    def replace(self, key: str, new_component: Union[SimComponentType, SimultanObject]):
        """
        Replace a existing Component or reference associated with a attribute (parameter name) with a new component
        or reference.

        There are six different cases:

        1. The old component is a sub-component and the new component is a SimComponent:
            the old component is exchanged with the new component

        2. The old component is a sub-component and the new component is a SimultanObject:
            the old component deleted and a reference to the SimultanObject is created

        3. The old component is a reference and the new component is a SimComponent:
            the reference to the old component is deleted a new sub-component is created

        4. The old component is a reference and the new component is a SimultanObject:
            the reference to the old component updated to reference the new component

        5. There is no old component and the new component is a SimComponent:
            A new sub-component is created

        6. There is no old component and the new component is a SimultanObject:
            A new reference is created

        References of other components to the old component are updated to the new component.

        :param key: the name of the attribute (parameter name)
        :param new_component: the new component to replace the old component
        """
        full_slot_name = self._slots.get(key, None)
        slot_name, slot_extension = parse_slot(full_slot_name)

        component = getattr(self, key)

        if component is None:
            if new_component is None:
                pass
            if isinstance(new_component, SimultanObject):
                self.add_component_reference(component=new_component,
                                             slot_name=slot_name,
                                             slot_extension=slot_extension)
            elif isinstance(new_component, SimComponent):
                self.add_sub_component(component=new_component,
                                       slot_name=slot_name,
                                       slot_extension=slot_extension,
                                       alter_slot=True)

        # old component is not None:
        else:
            idx = self.slot_components.index[np.array([x.Id for x in self.slot_components['comp']]) == component.Id][0]

            references = []
            for ref_comp in set(component._wrapped_obj.ReferencedBy):
                ref = next((x for x in ref_comp.ReferencedComponents.Items if x.Reference == component._wrapped_obj), None)
                references.append(ref)

            new_sim_component = new_component

            # old component is sub component
            if self.slot_components.iloc[idx].comp_type == 0:
                if isinstance(new_component, SimultanObject):
                    if new_component is not None:
                        new_sim_component = new_component._wrapped_obj
                        # add reference to component:
                        self.add_component_reference(component=new_component,
                                                     slot_name=slot_name,
                                                     slot_extension=slot_extension)
                    self.remove_sub_component(component)

                elif isinstance(new_component, SimComponent):
                    if new_component is not None:
                        comp_ref = next((x for x in self._wrapped_obj.Components.Items
                                    if ((x.Slot.SlotBase.Base == slot_name) & (x.Slot.SlotExtension == slot_extension)))
                                   , None)
                        new_component.CurrentSlot = comp_ref.Slot.SlotBase
                        comp_ref.Component = new_component

                elif new_component is None:
                    self.remove_sub_component(component)

            # old component is reference
            elif self.slot_components.iloc[idx].comp_type == 1:
                ref = next((x for x in self._wrapped_obj.ReferencedComponents.Items
                            if ((x.ReferenceFunction.SlotBase.Base == slot_name) & (
                        x.ReferenceFunction.SlotExtension == slot_extension))), None)
                if isinstance(new_component, SimultanObject):
                    new_sim_component = new_component._wrapped_obj
                    ref.Reference = new_component._wrapped_obj
                elif isinstance(new_component, SimComponent):
                    self.remove_component_reference(component)
                    self.add_sub_component(new_component)
                elif new_component is None:
                    self.remove_component_reference(component)

            for ref in references:
                ref.Reference = new_sim_component

        # clear the cache of the attributes getter:
        getattr(self.__class__, key).fget.cache_clear()

        # update the 'slot_components'
        self.slot_components = self.get_slot_components()


class List(UserList):

    _create_all = False     # if true all properties are evaluated to create python objects when initialized

    def __init__(self, *args, **kwargs):
        self._wrapped_obj = kwargs.get('wrapped_obj', None)
        self._contained_components = kwargs.get('contained_components', None)
        self._contained_parameters = kwargs.get('contained_parameters', None)
        self._template_parser = kwargs.get('template_parser', None)
        self._data_model_id = kwargs.get('data_model_id', None)

        if self._create_all:
            _ = self.contained_components
            _ = self.contained_parameters

    @property
    def data(self):
        if self._wrapped_obj is None:
            return
        components = [x.Component for x in self._wrapped_obj.Components.Items]
        try:
            indices = sort_component_list(self._wrapped_obj.Components.Items)
            return [x for _, x in sorted(zip(indices, components))]
        except TypeError as e:
            logger.warning(f'Could not sort list {components}:\n{e}')
            return components

    def __getitem__(self, i):
        if isinstance(i, slice):
            if self._template_parser is None:
                return self.__class__(self.data[i])
            return [self._template_parser.create_python_object(x) for x in self.__class__(self.data[i])]
        else:
            if self._template_parser is None:
                return self.data[i]
            return self._template_parser.create_python_object(self.data[i])

    def __repr__(self):
        return f'{self.name}: ' + repr(list(self.data))

    @property
    def contained_components(self):
        if self._contained_components is None:
            if self._wrapped_obj is not None:
                self._contained_components = [self._template_parser.create_python_object(x.Component) for x in self._wrapped_obj.Components.Items]
        return self._contained_components

    @property
    def contained_parameters(self):
        if self._contained_parameters is None:
            if self._wrapped_obj is not None:
                self._contained_parameters = {x.Name: x.get_ValueCurrent() for x in self._wrapped_obj.Parameters.Items}
        return self._contained_parameters

    @property
    def name(self):
        if self._wrapped_obj is not None:
            if hasattr(self._wrapped_obj, 'Name'):
                return self._wrapped_obj.Name

    @name.setter
    def name(self, value):
        if self._wrapped_obj is not None:
            self._wrapped_obj.Name = value


class ReferenceList(UserList):

    def __init__(self, *args, **kwargs):
        self._wrapped_obj = kwargs.get('wrapped_obj', None)
        self._contained_components = kwargs.get('contained_components', None)
        self._contained_parameters = kwargs.get('contained_parameters', None)
        self._template_parser = kwargs.get('template_parser', None)
        self._data_model_id = kwargs.get('data_model_id', None)

    def __getitem__(self, i):
        if isinstance(i, slice):
            if self._template_parser is None:
                return self.__class__(self.data[i])
            return [self._template_parser.create_python_object(x) for x in self.__class__(self.data[i])]
        else:
            if self._template_parser is None:
                return self.data[i]
            return self._template_parser.create_python_object(self.data[i])

    def __repr__(self):
        return f'{self.name}: ' + repr(list(self.data))

    @property
    def data(self):
        if self._wrapped_obj is None:
            return

        ref_components = [x.Reference for x in self._wrapped_obj.ReferencedComponents.Items]
        try:
            indices = sort_references(self._wrapped_obj.ReferencedComponents.Items)
            return [x for _, x in sorted(zip(indices, ref_components))]
        except TypeError as e:
            logger.warning(f'Could not sort list {ref_components}:\n{e}')
            return ref_components

    @property
    def contained_components(self):
        if self._contained_components is None:
            if self._wrapped_obj is not None:
                self._contained_components = [self._template_parser.create_python_object(x) for x in
                                              self._wrapped_obj.Components]
        return self._contained_components

    @property
    def contained_parameters(self):
        if self._contained_parameters is None:
            if self._wrapped_obj is not None:
                self._contained_parameters = {x.Name: x.get_ValueCurrent() for x in
                                              self._wrapped_obj.Parameters.Items}
        return self._contained_parameters

    @property
    def name(self):
        if self._wrapped_obj is not None:
            if hasattr(self._wrapped_obj, 'Name'):
                return self._wrapped_obj.Name

    @name.setter
    def name(self, value):
        if self._wrapped_obj is not None:
            self._wrapped_obj.Name = value


class ValueField(pd.DataFrame):

    _metadata = pd.DataFrame._metadata + ['_wrapped_obj', '_template_parser', '_data_model_id']

    def __init__(self, *args, **kwargs):

        # self._wrapped_obj = kwargs.get('wrapped_obj')
        self._wrapped_obj = kwargs.pop('wrapped_obj', None)
        self._template_parser = kwargs.get('template_parser', None)
        self._data_model_id = kwargs.get('data_model_id', None)

        row_headers = [x.Name for x in self._wrapped_obj.ValueField.RowHeaders.Items]
        column_headers = [x.Name for x in self._wrapped_obj.ValueField.ColumnHeaders.Items]
        values = self._wrapped_obj.ValueField.Values

        super().__init__(dict(zip(column_headers, np.array(values).T)),  index=row_headers)

    @property
    def base_class_view(self):
        # use this to view the base class, needed for debugging in some IDEs.
        return pd.DataFrame(self)


class BuildInFace(SimultanObject):

    def __init__(self, *args, **kwargs):
        SimultanObject.__init__(self, *args, **kwargs)

        self._geo_instances = kwargs.get('geo_instances', None)
        self._boundaries = kwargs.get('boundaries', None)

    @property
    def geo_ids(self):
        if self.geo_instances is not None:
            return [x.Id for x in self.geo_instances]

    @property
    def area(self):
        # idx = next((i for i, x in enumerate(self._wrapped_obj.ContainedParameters.Items) if x.Name == 'A'), None)
        # param = self._wrapped_obj.ContainedParameters.Items[idx]
        # return param.get_ValueCurrent()
        return self.get_param('A')

    @property
    def geo_ids(self):
        geo_ids = []
        for item in self._wrapped_obj.Instances.Items:

            geom_placement = next(x for x in item.Placements.Items if isinstance(x, SimInstancePlacementGeometry))

            file_id = geom_placement.FileId
            geometry_id = geom_placement.GeometryId

            geo_ids.append({'FileId': file_id,
                           'GeometryId': geometry_id})
        return geo_ids

    @property
    def geo_instances(self):
        if self._geo_instances is None:
            self._geo_instances = self.get_geo_instances()
        return self._geo_instances

    @geo_instances.setter
    def geo_instances(self, value):
        self._geo_instances = value

    @property
    def boundaries(self):
        if self._boundaries is None:
            if self.geo_instances is not None:
                self._boundaries = [x.boundary for x in self.geo_instances]
        return self._boundaries

    @property
    def construction(self):
        obj = next((x.Reference for x in self._wrapped_obj.ReferencedComponents.Items if x.ReferenceFunction.SlotFull == 'Aufbau_0AG0'), None)
        return self._template_parser.create_python_object(obj)

    def get_geo_instances(self):
        try:
            geo_instances = []

            for item in self._wrapped_obj.Instances.Items:

                if item.Placements.Items.__len__() == 0:
                    continue

                geom_placement = next(x for x in item.Placements.Items if isinstance(x, SimInstancePlacementGeometry))

                file_id = geom_placement.FileId
                geometry_id = geom_placement.GeometryId

                # geo_model = self._template_parser.data_models[self._data_model_id].get_typed_model_by_file_id(file_id)
                geo_model = self._template_parser.typed_geo_models[file_id]
                geo_instance = geo_model.get_face_by_id(geometry_id)

                # geo_instance = self._template_parser.data_models[self._data_model_id].typed_geo_models[file_id].get_face_by_id(geometry_id)

                geo_instances.append(geo_instance)

            return geo_instances
        except Exception as e:
            logger.error(f'BuildInFace {self.name}; {self.id}: Error while getting geo instances: {e}')
            return []

    def associate_with_geometry(self, face):

        try:
            self.InstanceType = SimInstanceType.Entity3D
        except Exception as e:
            raise e

        exch = face.geometry_model.template_parser.current_data_model.exch
        exch.Associate(self, face._wrapped_obj)


class BuildInVolume(SimultanObject):

    def __init__(self, *args, **kwargs):
        """
        Default python class for SIMULTAN 'Geometrische_Volumina' Slot
        Grundflächen und Rauminhalte nach DIN 277

        @keyword geo_instances: List of geometric volume instances of type geo_default_types.GeometricVolume
        @keyword surfaces: List of
        """
        SimultanObject.__init__(self, *args, **kwargs)

        self._geo_instances = kwargs.get('geo_instances', None)
        self._surfaces = kwargs.get('surfaces', None)
        self._volumes = kwargs.get('volumes', None)

    @property
    def volumes(self):
        return self.geo_instances

    @property
    def geo_faces(self):
        """
        Faces of the Geometry-Model
        :return: list with faces of type geo_defaut_types.GeometricFace
        """
        faces = []
        try:
            [faces.extend(x.faces) for x in self.geo_instances if x is not None]
        except AttributeError as e:
            raise e
        return faces

    @property
    def face_components(self):
        return None

    @property
    def geo_instances(self):
        if self._geo_instances is None:
            self._geo_instances = self.get_geo_instances()
        return self._geo_instances

    @geo_instances.setter
    def geo_instances(self, value):
        self._geo_instances = value

    @property
    @lru_cache(maxsize=None)
    def v_a(self):
        return next((x.ValueCurrent for x in self._wrapped_obj.Parameters.Items if x.Name == 'Vᴀ'), None)

    @property
    @lru_cache(maxsize=None)
    def v_nri(self):
        """
        Netto-Rauminhalt NRI: Anteil des Brutto-Rauminhalts (BRI), der das Volumen über der Netto-Raumfläche (NRF) umfasst; DIN 277
        :return:
        """
        return next((x.ValueCurrent for x in self._wrapped_obj.Parameters.Items if x.Name == 'Vɴʀɪ'), None)

    @property
    @lru_cache(maxsize=None)
    def v_bri(self):
        """
        Brutto-Rauminhalt BRI: gesamtes Volumen eines Bauwerks oder eines Geschosses, das sich in Netto-Rauminhalt (NRI) und
        Konstruktions-Rauminhalt (KRI) gliedert; DIN 277
        :return:
        """
        return next((x.ValueCurrent for x in self._wrapped_obj.Parameters.Items if x.Name == 'Vᴃʀɪ'), None)

    @property
    @lru_cache(maxsize=None)
    def a_bgf(self):
        """
        Brutto-Grundfläche; gesamte Grundfläche eines Bauwerks oder eines Geschosses, die sich in Netto-Raumfläche (NRF)
        und Konstruktions-Grundfläche (KGF) gliedert; DIN 277
        :return:
        """
        return next((x.ValueCurrent for x in self._wrapped_obj.Parameters.Items if x.Name == 'Aᴃɢꜰ'), None)

    @property
    @lru_cache(maxsize=None)
    def a_ngf(self):
        """
        Netto-Grundfläche; DIN 277
        :return:
        """
        return next((x.ValueCurrent for x in self._wrapped_obj.Parameters.Items if x.Name == 'Aɴɢꜰ'), None)

    def get_geo_instances(self):

        geo_instances = []

        for item in self._wrapped_obj.Instances.Items:

            try:
                geom_placement = next(x for x in item.Placements.Items if isinstance(x, SimInstancePlacementGeometry))
            except Exception as e:
                logger.error(f'{self.__class__.__name__} {self.name} {self.id}: Could not find Geometry Placement for GeometryInstance {item}')
                if continue_on_error:
                    continue
                else:
                    raise e

            file_id = geom_placement.FileId
            geometry_id = geom_placement.GeometryId

            # geo_model = self._template_parser.data_models[self._data_model_id].get_typed_model_by_file_id(file_id)
            geo_model = self._template_parser.typed_geo_models[file_id]
            geo_instance = geo_model.get_zone_by_id(geometry_id)

            if geo_instance is None:
                logger.error(f'{self.__class__.__name__} {self.name} {self.id}: Geometry Instance with file id: {file_id}, geometry id: {geometry_id} not found')
                if not continue_on_error:
                    raise KeyError(f'{self.__class__.__name__} {self.name} {self.id}: Geometry Instance with file id: {file_id}, geometry id: {geometry_id} not found')

            else:
                geo_instances.append(geo_instance)

        return geo_instances


class BuildInZone(SimultanObject):

    def __init__(self, *args, **kwargs):
        """

        """
        SimultanObject.__init__(self, *args, **kwargs)

    @property
    @lru_cache(maxsize=None)
    def volumes(self):
        return [x for x in self.contained_components if x._wrapped_obj.FitsInSlots[0] == 'Geometrische_Volumina']

    @property
    @lru_cache(maxsize=None)
    def faces(self):
        return [x for x in self.contained_components if x._wrapped_obj.FitsInSlots[0] == 'Geometrische_Flächen']


class BaseBuildInConstruction(SimultanObject):

    def __init__(self, *args, **kwargs):
        """
        Default implementation for a wall construction (InstanceType == InstanceType.ALIGNED_WITH).
        """
        SimultanObject.__init__(self, *args, **kwargs)
        self.check_type()

    @property
    def is_window(self):
        if self.get_param_index('gVergl') is None:
            return False
        else:
            return True

    def check_type(self):
        if self.is_window:
            object.__setattr__(self, '__class__', BuildInWindowConstruction)
        else:
            object.__setattr__(self, '__class__', BuildInWallConstruction)


class BuildInWallConstruction(BaseBuildInConstruction):

    def __init__(self, *args, **kwargs):
        """
        Default implementation for a wall construction (InstanceType == InstanceType.ALIGNED_WITH).
        """
        BaseBuildInConstruction.__init__(self, *args, **kwargs)

    @property
    def total_thickness(self):
        idx = self.get_param_index('DickeGes')
        # idx = next((i for i, x in enumerate(self._wrapped_obj.ContainedParameters.Items) if x.Name == 'DickeGes'), None)
        param = self._wrapped_obj.Parameters.Items[idx]
        return param.get_ValueCurrent()

    @property
    def layers(self):
        layers = [self._template_parser.create_python_object(x.Component, template_name='BuildInMaterialLayer') for x in self._wrapped_obj.Components.Items]

        # sort the layers by their slot_extension:
        # slot_extensions = [x.slot_extension for x in layers]
        slot_extensions = [int(x.Slot.SlotExtension) for x in self._wrapped_obj.Components.Items]
        # test with rotated
        # layers = layers[1:] + layers[:1]

        return [x for _, x in sorted(zip(slot_extensions, layers))]


class BuildInWindowConstruction(BaseBuildInConstruction):

    def __init__(self, *args, **kwargs):
        """
        Default implementation for a window construction (InstanceType == InstanceType.ALIGNED_WITH).
        """
        BaseBuildInConstruction.__init__(self, *args, **kwargs)

    @property
    def d(self):
        """
        layer thickness in m
        :param
        """
        try:
            return self.get_param('d')
        except AttributeError:
            return self.get_param('Dicke')

    @d.setter
    def d(self, value):
        # idx = self.get_param_index('d')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('d', value)

    @property
    def eps(self):
        """
        eps
        :param
        """
        return self.get_param('eps')

    @eps.setter
    def eps(self, value):
        # idx = self.get_param_index('eps')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('eps', value)

    @property
    def gVergl(self):
        """
        Gesamtenergiedurchlassgrad
        :param
        """
        return self.get_param('gVergl')

    @gVergl.setter
    def gVergl(self, value):
        # idx = self.get_param_index('gVergl')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('gVergl', value)

    @property
    def curtain_position(self):
        """

        :param
        """
        return self.get_param('LageBehang')

    @curtain_position.setter
    def curtain_position(self, value):
        # idx = self.get_param_index('LageBehang')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('LageBehang', value)

    @property
    def openability(self):
        """
        openability of a window; 1=vollständig  0.1=kippen
        :param
        """
        return self.get_param('Öffenbarkeit')

    @openability.setter
    def openability(self, value):
        # idx = self.get_param_index('Öffenbarkeit')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('Öffenbarkeit', value)

    @property
    def frame_width(self):
        """
        frame width in m
        :param
        """
        return self.get_param('Rahmenbreite')

    @frame_width.setter
    def frame_width(self, value):
        # idx = self.get_param_index('Rahmenbreite')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('Rahmenbreite', value)

    @property
    def u_f(self):
        """
        U-Wert Rahmen in W/m²K
        :param
        """
        return self.get_param('Uf')

    @u_f.setter
    def u_f(self, value):
        # idx = self.get_param_index('Uf')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('Uf', value)

    @property
    def u_g(self):
        """
        U-Wert Glas in W/m²K
        :param
        """
        return self.get_param('Ug')

    @u_g.setter
    def u_g(self, value):
        self.set_param('Ug', value)

    @property
    def alpha_curtain(self):
        """
        Absorption Behang
        :param
        """
        return self.get_param('αeB')

    @alpha_curtain.setter
    def alpha_curtain(self, value):
        self.set_param('αeB', value)

    @property
    def rho_curtain(self):
        """
        Reflexion Behang
        :param
        """
        return self.get_param('ρeB')

    @rho_curtain.setter
    def rho_curtain(self, value):
        self.set_param('ρeB', value)

    @property
    def tau_curtain(self):
        """
        Transmission Behang
        :param
        """
        return self.get_param('τeB')

    @tau_curtain.setter
    def tau_curtain(self, value):
        self.set_param('τeB', value)

    @property
    def psi(self):
        """
        Transmission Behang
        :param
        """
        return self.get_param('ψ')

    @psi.setter
    def psi(self, value):
        self.set_param('ψ', value)


class BuildInMaterialLayer(SimultanObject):

    def __init__(self, *args, **kwargs):
        """
        Default implementation for a material layer of a wall construction.
        """
        SimultanObject.__init__(self, *args, **kwargs)
        print('done')

    @property
    def absorption_rate(self):
        """
        absorption rate (accoustic?)
        :param
        """
        return self.get_param('aGr')

    @absorption_rate.setter
    def absorption_rate(self, value):
        # idx = self.get_param_index('aGr')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('aGr', value)

    @property
    def c(self):
        """
        specific heat capacity in J/kgK
        :param
        """
        return self.get_param('c')

    @c.setter
    def c(self, value):
        # idx = self.get_param_index('c')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('c', value)

    @property
    def d(self):
        """
        layer thickness in m
        :param
        """
        return self.get_param('d')

    @d.setter
    def d(self, value):
        # idx = self.get_param_index('d')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('d', value)

    @property
    def w20(self):
        """
        Ausgleichsfeuchte bei rli 20%
        :param
        """
        return self.get_param('w20')

    @w20.setter
    def w20(self, value):
        # idx = self.get_param_index('w20')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('w20', value)

    @property
    def w80(self):
        """
        Ausgleichsfeuchte bei rli 80%
        :param
        """
        return self.get_param('w80')

    @w80.setter
    def w80(self, value):
        # idx = self.get_param_index('w80')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('w80', value)

    @property
    def heat_conductivity(self):
        """
        heat conductivity in W/mK
        :param
        """
        return self.get_param('λ')

    @heat_conductivity.setter
    def heat_conductivity(self, value):
        # idx = self.get_param_index('λ')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('λ', value)

    @property
    def mu(self):
        """
        Wasserdampfdiffusionswiderstand
        :param
        """
        return self.get_param('μ')

    @mu.setter
    def mu(self, value):
        # idx = self.get_param_index('μ')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('μ', value)

    @property
    def rho(self):
        """
        Rohdichte in kg/m³
        :param
        """
        return self.get_param('ρ')

    @rho.setter
    def rho(self, value):
        # idx = self.get_param_index('ρ')
        # self._wrapped_obj.ContainedParameters.Items[idx].set_ValueCurrent(value)
        self.set_param('ρ', value)


def create_component(name=None, parameters=None, slot=ComponentUtils.COMP_SLOT_UNDEFINED):
    """
    Creates a new SIMULTAN Component, optional add parameters
    :param name: Name of the component
    :param parameters: dictionary with parameters: {<parameter_name>: [<value>, <unit>]};
    Example: {'length': [10, 'm'], 'weight': [13.0, 'kg']}
    :param slot: Component Slot, sting or ComponentUtils
    :return: SimComponent
    """
    new_comp = SimComponent()

    if name is not None:
        new_comp.Name = 'Test'

    # set parameters
    if parameters is not None:
        for key, value in parameters.items():
            new_comp.Parameters.Add(create_parameter(name=key,
                                                     value=value[0],
                                                     unit=value[1]))
    # set slot:
    new_comp.CurrentSlot = SimSlotBase(slot)
    return new_comp



from ruamel.yaml import YAML, yaml_object, add_representer
from .config import yaml
import io
import colorlog
from uuid import uuid4
import typing

from yamale.readers.yaml_reader import _parsers
import yamale


from collections import UserList
from .default_types import SimultanObject
from .default_types import List as SimultanList
from .default_types import ValueField, BuildInFace, BuildInVolume, BuildInZone, BaseBuildInConstruction, BuildInMaterialLayer, ReferenceList
from .geo_default_types import geometry_types
from .geometry import GeometryModel
from .utils import parse_slot

from functools import lru_cache

from System.Linq import Enumerable
from ParameterStructure.Values import SimMultiValuePointer
from .ParameterStructure.Components import SimComponent

from typing import List as TypeList
from typing import NewType, Union

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 importlib_resources.
    import importlib_resources as pkg_resources

from . import resources

with pkg_resources.path(resources, 'schema.yaml') as r_path:
    schema_file = str(r_path)


logger = colorlog.getLogger('PySimultan')


template_classes = {}


# create the class for the templates
@yaml_object(yaml)
class Template(object):

    yaml_tag = u'!Template:'

    def __init__(self, *args, **kwargs):
        """ Template class to define how SIMULTAN SimComponents are imported and parsed

        :param args:
        :param kwargs:

        @keyword template_name: Name of the Template. This name is the name of the created Python class and used for
        TYPE matching the SIMULTAN SimComponent
        @keyword template_id: ID of the template; int
        @keyword inherits_from: name of the python template or the template instance this template inherits from
        @keyword content: List of attribute names
        @keyword documentation: Entry to document the template; str
        @keyword units: units of the content. Dictionary
        @keyword types: Dictionary with the type of the content. if the type is 'str' get_TextValue() is returned, param.get_ValueCurrent() otherwise
        @keyword slots: Dictionary with the name of the content and the name of the slot where it is expected. The expected slot is a string
        @keyword default_slot: Default slot name for the Simultan component
        with the slot name and the extension, seperated by a blank
        """

        self.template_name = kwargs.get('template_name', None)
        self.template_id = kwargs.get('template_id', None)

        self.inherits_from = kwargs.get('inherits_from', None)

        self.content = kwargs.get('content', [])
        self.documentation = kwargs.get('documentation', None)
        self.units = kwargs.get('units', {})
        self.types = kwargs.get('types', {})

        self.slots = kwargs.get('slots', {})
        self.synonyms = kwargs.get('synonyms', {})      # name of the attribute in the python class

        self.template_parser = kwargs.get('template_parser', None)
        self.template_class = kwargs.get('template_class', None)

        self.default_slot = kwargs.get('default_slot', 'Element')

    def write(self, filename: str = None):
        """
        write the template as .yml to a file
        :param filename: name and path of the file to save; Example: C:\SIMULTAN\my_template.yml
        :return: None
        """
        if filename is not None:
            yaml.dump([self], open(filename, mode='w'))
        else:
            f = io.StringIO()
            yaml.dump([self], f)
            return f.getvalue()

    def create_template_class(self, template_parser, template_classes):
        """
        Creates a template class for this template.
        :param template_parser: TemplateParser instance
        :param template_classes: list of template classes to inherit from (e.g. default classes)
        :return: TemplateClass
        """

        if (self.template_parser is template_parser) and (self.template_class is not None):
            return self.template_class

        self.template_parser = template_parser

        # for inheritance:
        # template classes must inherit from SimultanObject, UserList or ValueField

        if self.inherits_from is not None:
            if self.inherits_from in self.template_parser.bases.keys():
                base = template_classes[self.inherits_from]

                simultan_base_list = [SimultanObject, UserList, ValueField]

                if any(x in simultan_base_list for x in base.__bases__):
                    bases = (base, )
                else:
                    bases = (SimultanObject, ) + (base, )
                    # bases = (base, ) + (SimultanObject, )

                # if SimultanObject in base.__bases__:
                #     bases = (base, )
                # elif UserList in base.__bases__:
                #     bases = (base, )
                # elif ValueField in base.__bases__:
                #     bases = (base, )
                # else:
                #     bases = (SimultanObject, ) + (base, )
            else:
                base = template_classes[self.inherits_from.template_name]
                bases = (base, )

            def new_init(self, *args, **kwargs):

                for i in range(self.__class__.__bases__.__len__()):
                    self.__class__.__bases__[i].__init__(self, *args, **kwargs)

        else:
            bases = (SimultanObject, )
            # create the class from the template

            def new_init(self, *args, **kwargs):

                self.__class__.__bases__[0].__init__(self, *args, **kwargs)

        new_class_dict = {'__init__': new_init,
                          '_template_name': self.template_name,
                          '_template_id': self.template_id,
                          '_documentation': self.documentation,
                          '_content': [cont for cont in self.content],
                          '_types': self.types,
                          '_units': self.units,
                          '_base': bases,
                          '_template_parser': self.template_parser,
                          '_slots': self.slots,
                          '_synonyms': self.synonyms,
                          '_default_slot': self.default_slot}

        new_class_dict.update(self.get_properties())
        new_class = type(self.template_name, bases, new_class_dict)

        self.template_class = new_class

        return new_class

    def get_properties(self):

        prop_dict = {}

        for prop in self.content:

            if prop in self.slots.keys():
                slot = self.slots[prop]
            else:
                slot = None

            syn = self.synonyms.get('prop', prop)

            prop_dict[syn] = add_properties(prop, syn, param_type=self.types.get(prop, None), slot=slot)

        return prop_dict

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['template_parser']
        del state['template_class']
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        self.template_parser = None
        self.template_class = None

    def __repr__(self):
        return f"Template '{self.template_name}': " + object.__repr__(self)


class TemplateParser(object):

    bases = {'Liste': SimultanList,
             'List': SimultanList,
             'ReferenceList': ReferenceList,
             'ValueField': ValueField,
             'Geometric Area': BuildInFace,
             'Geometrische_Flächen': BuildInFace,
             'Geometric Volume': BuildInVolume,
             'Geometrische_Volumina': BuildInVolume,
             'BuildInZone': BuildInZone,
             'BuildInConstruction': BaseBuildInConstruction,
             'BuildInMaterialLayer': BuildInMaterialLayer}

    geo_bases = {'base': geometry_types.base,
                 'layer': geometry_types.layer,
                 'vertex': geometry_types.vertex,
                 'edge': geometry_types.edge,
                 'edge_loop': geometry_types.edge_loop,
                 'face': geometry_types.face,
                 'volume': geometry_types.volume,
                 }

    _create_all = False

    def __init__(self, *args, **kwargs):

        """
        Class which handles templates. This class generates python objects from the SIMULTAN data model for templates.

        @keyword templates: list of templates; if templates is None and template_filepath is defined,
        templates are automatically loaded
        @keyword template_filepath: filepath to the template (*.yml)

        """

        self._current_data_model = None

        self.id = uuid4()

        self._templates = kwargs.get('templates', None)
        self.template_filepath = kwargs.get('template_filepath', None)

        self._template_classes = kwargs.get('template_classes', None)

        self.data_models = kwargs.get('data_models', {})

        self._typed_geo_models = None

        template_classes[self.id] = self._template_classes

        self.current_data_model = kwargs.get('current_data_model', None)

    @property
    def current_data_model(self):
        return self._current_data_model

    @current_data_model.setter
    def current_data_model(self, value):
        self._current_data_model = value
        if self._current_data_model is None:
            return

        if self._current_data_model not in self.data_models:
            self.data_models[self._current_data_model.id] = self._current_data_model

    @property
    def templates(self):
        if self.template_filepath is None:
            return None

        if self._templates is None:
            self._templates = self.load_templates_from_file()
        return self._templates

    @property
    def template_classes(self):
        if self._template_classes is None:
            self.create_template_classes()
        return self._template_classes

    @property
    def typed_geo_models(self):
        if self._typed_geo_models is None:
            self._typed_geo_models = self.get_typed_geo_models()
        return self._typed_geo_models

    def get_model_by_file_id(self, id):
        return self.current_data_model.models_dict[id]

    def get_typed_model_by_file_id(self, id):
        return self.typed_geo_models[id]

    def load_templates_from_file(self, filepath=None):
        """
        Load templates from file
        :param filepath: filepath of the file (str)
        :return: templates; list of templates
        """
        if filepath is None:
            filepath = self.template_filepath

        if filepath is None:
            templates = []
        else:
            with open(filepath, mode='r', encoding="utf-8") as f_obj:
                templates = yaml.load(f_obj)

        return templates

    def create_template_classes(self) -> typing.Dict[str, str]:

        logger.info('\n\nCreating template-classes:\n-------------------------------------')
        template_classes = {**self.bases}

        if self.templates is not None:
            logger.info(f'found {self.templates.__len__()} templates')

            for template in self.templates:

                new_class = template.create_template_class(self, template_classes)
                template_classes[template.template_name] = new_class

                logger.info(f'Created template class {template.template_name}. Inherits from {template.inherits_from}')

        else:
            logger.info(f'No template file defined')

        self._template_classes = template_classes
        logger.info('template-class creation finished\n\n')
        return self._template_classes

    @lru_cache(maxsize=None)
    def create_python_object(self, sim_component, template_name=None):

        # get the template or slot
        # template_name = None
        if hasattr(sim_component, 'Parameters'):
            t_name = next((x.TextValue for x in sim_component.Parameters.Items if x.Name == 'TYPE'), None)
            if t_name is not None:
                template_name = t_name

        if (template_name is None) and (hasattr(sim_component, 'get_CurrentSlot')):
            template_name = sim_component.CurrentSlot.Base

        # if no template and no slot could be found return the plain SimComponent

        if template_name is None:
            if isinstance(sim_component, SimMultiValuePointer):
                template_name = 'ValueField'

        if template_name is None:
            if self._create_all:
                if hasattr(SimComponent, 'ContainedSimComponentsAsList'):
                    _ = [self.create_python_object(x) for x in sim_component.Components]
                if hasattr(SimComponent, 'ReferencedSimComponents'):
                    _ = [self.create_python_object(x) for x in sim_component.ReferencedComponents.Items]
            return sim_component

        if template_name not in self.template_classes.keys():

            # create new class for the template / slot
            base = SimultanObject
            bases = (base, )

            def new__init(self, *args, **kwargs):
                self.__class__.__bases__[0].__init__(self, *args, **kwargs)

            new_class_dict = {'__init__': new__init,
                              '_template_name': template_name,
                              '_template_id': None,
                              '_documentation': None,
                              '_content': None,
                              '_types': None,
                              '_units': None,
                              '_base': bases,
                              '_template_parser': self,
                              '_slots': {}}

            new_class = type(template_name, bases, new_class_dict)
            self.template_classes[template_name] = new_class

        template_class = self.template_classes[template_name]

        # init new instance
        template_class._template_parser = self
        new_instance = template_class(wrapped_obj=sim_component,
                                      template_parser=self,
                                      data_model_id=self._current_data_model.id)

        # new_instance._template_parser = self
        return new_instance

    def get_typed_geo_models(self):
        typed_models_dict = {}
        for key, model in self.current_data_model.models_dict.items():
            if model is None:
                typed_models_dict[key] = None
                continue
            typed_models_dict[key] = GeometryModel(template_parser=self,
                                                   wrapped_obj=model,
                                                   geo_types=self.geo_bases)
            # models.append(GeometryModel(wrapped_obj=model))
        return typed_models_dict

    def add_new_geometry_model(self, file_name: str, model_name: str = None):

        model, resource = self.current_data_model.add_new_geometry_model(file_name, model_name, return_resource=True)

        return self.add_typed_geo_model(model, resource.Key)

    def add_typed_geo_model(self, model, key):
        """
        Add a geometry model as typed model
        :param model: ParameterStructure.Assets.ContainedResourceFileEntry
        :param key: resource key
        """
        if self._typed_geo_models is None:
            self._typed_geo_models = {}

        self._typed_geo_models[key] = GeometryModel(template_parser=self,
                                                    wrapped_obj=model,
                                                    geo_types=self.geo_bases)

        return self._typed_geo_models[key]

    def get_geo_components(self, geometry):
        return Enumerable.ToList[SimComponent](self.current_data_model.exch.GetComponents(geometry))

    def get_py_geo_components(self, geometry, template_name=None):
        components = Enumerable.ToList[SimComponent](self.current_data_model.exch.GetComponents(geometry))
        return [self.create_python_object(x, template_name=template_name) for x in components]


def add_properties(prop_name: str, syn_name: str, param_type: str, slot=None) -> property:
    """
    create property for a class
    :param prop_name: name of the property (str)
    :param prop_name: name of the synonym (str)
    :param param_type: type of the property; if type == 'str': param.get_TextValue(); else: param.get_ValueCurrent() is returned
    :param slot: slot; default is None
    :return: property
    """

    @lru_cache()
    def getx(self):
        obj = None

        idx = next((i for i, x in enumerate(self._wrapped_obj.Parameters.Items) if x.Name == prop_name), None)
        if idx is not None:
            param = self._wrapped_obj.Parameters.Items[idx]
            obj = param.get_MultiValuePointer()

            if obj is None:
                if param_type == 'str':
                    obj = param.get_TextValue()
                else:
                    obj = param.get_ValueCurrent()

        if obj is None:
            full_slot_name = self._slots.get(prop_name, None)
            slot_name, slot_extension = parse_slot(full_slot_name)

            if full_slot_name is not None:
                obj = next((x.Component for x in self._wrapped_obj.Components.Items
                            if (x.Slot.SlotBase.Base == slot_name) & (x.Slot.SlotExtension == slot_extension)),
                           None)

                if obj is None:
                    obj = next((x.get_Reference() for x in self._wrapped_obj.ReferencedComponents.Items
                                if ' '.join([str(x.ReferenceFunction.SlotBase),
                                             x.ReferenceFunction.SlotExtension]) == full_slot_name),
                               None)

        return self._template_parser.create_python_object(obj)

    def setx(self, value: Union[int, float, SimComponent, SimultanObject]):
        getx.cache_clear()
        idx = next((i for i, x in enumerate(self._wrapped_obj.Parameters.Items) if x.Name == prop_name), None)

        if isinstance(value, (SimComponent, SimultanObject)):
            # set reference or subcomponent

            if prop_name not in self._slots.keys():
                raise KeyError(f'{prop_name} not defined as slot for class {self.__class__.__name__}')

            full_slot_name = self._slots.get(prop_name, None)
            slot_name, slot_extension = parse_slot(full_slot_name)

            # check if slot exists:
            idx = self.slot_components.index[
                (self.slot_components.slot == slot_name) & (self.slot_components.slot_extension == slot_extension)]

            if idx.__len__() == 1:
                # if subcomponent:
                if self.slot_components.iloc[idx].comp_type.values == 0:
                    if isinstance(value, SimultanObject):
                        pass

                    ref = next((x for x in self._wrapped_obj.Components.Items
                                if ((x.Slot.SlotBase.Base == slot_name) & (
                                x.Slot.SlotExtension == slot_extension))), None)

                    if isinstance(value, SimComponent):
                        ref.Component = value
                    elif isinstance(value, SimultanObject):
                        ref.Component = value._wrapped_obj

                # if reference:
                elif self.slot_components.iloc[idx].comp_type.values == 1:
                    ref = next((x for x in self._wrapped_obj.ReferencedComponents.Items
                                if ((x.ReferenceFunction.SlotBase.Base == slot_name) & (
                                x.ReferenceFunction.SlotExtension == slot_extension))), None)

                    if isinstance(value, SimComponent):
                        # delete reference
                        self.remove_component_reference(ref.reference)
                        # add the SimComponent as sub-component
                        self.add_sub_component(component=value,
                                               slot_name=slot_name,
                                               slot_extension=slot_extension,
                                               alter_slot=True)
                    else:
                        ref.Reference = value._wrapped_obj

            # index does not exist
            elif idx.__len__() == 0:
                if isinstance(value, SimultanObject):
                    self.add_component_reference(component=value._wrapped_obj,
                                                 slot_name=slot_name,
                                                 slot_extension=slot_extension)
                elif isinstance(value, SimComponent):
                    self.add_sub_component(component=value, slot_name=slot_name, slot_extension=slot_extension)

        elif param_type == 'str':
            #todo: implementation for referenced parameters
            if self._wrapped_obj.Parameters.Items[2].GetReferencedParameter() != self._wrapped_obj.Parameters.Items[idx]:
                raise NotImplementedError(f'{self.name}: Param {prop_name} is a referenced Parameter. Setting referenced parameters is not yet implemented')
            self._wrapped_obj.Parameters.Items[idx].set_TextValue(value)
        elif isinstance(value, (int, float)):
            # todo: implementation for referenced parameters
            if self._wrapped_obj.Parameters.Items[2].GetReferencedParameter() != self._wrapped_obj.Parameters.Items[idx]:
                raise NotImplementedError(f'{self.name}: Param {prop_name} is a referenced Parameter. Setting referenced parameters is not yet implemented')
            self._wrapped_obj.Parameters.Items[idx].set_ValueCurrent(float(value))
        else:
            raise TypeError(
                f'{self.name}: Error setting parameter {prop_name} to {type(value)}. Supported types are only: int float str SimComponent SimultanObject.')

    def delx(self):
        logger.warning('delete method not implemented')
        raise NotImplementedError

    return property(getx, setx, delx, f"automatic created property")


def _my_loader(f):
    import yaml

    description = []

    def any_constructor(loader, tag_suffix, node):
        if isinstance(node, yaml.MappingNode):
            return loader.construct_mapping(node)
        if isinstance(node, yaml.SequenceNode):
            return loader.construct_sequence(node)
        return loader.construct_scalar(node)

    yaml.add_multi_constructor('', any_constructor, Loader=yaml.SafeLoader)
    data = list(yaml.load_all(f, Loader=yaml.SafeLoader))
    return data


_parsers['_my_loader'] = _my_loader


class TemplateValidator(object):

    def __init__(self, *args, **kwargs):

        self.schema = yamale.make_schema(schema_file, parser='ruamel')

    def validate(self, yaml_file):
        data = yamale.make_data(yaml_file, parser='_my_loader')
        yamale.validate(self.schema, data)



from distutils.core import setup, Extension
from pathlib import Path

import setuptools

project_dir = Path(__file__).parent

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setuptools.setup(
    name="PySimultan",
    version="0.1.36",
    description="Python Package to import and work with the SIMULTAN Data model",
    # Allow UTF-8 characters in README with encoding argument.
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=["python"],
    author="Max Buehler",
    url="https://github.com/bph-tuwien/PySimultan",
    packages=["PySimultan"],
    package_dir={"PySimultan": "src/PySimultan"},
    package_data={'PySimultan': ['resources/*']},
    # pip 9.0+ will inspect this field when installing to help users install a
    # compatible version of the library for their Python version.
    python_requires=">=3.7",
    setup_requires=["wheel"],
    # There are some peculiarities on how to include package data for source
    # distributions using setuptools. You also need to add entries for package
    # data to MANIFEST.in.
    # See https://stackoverflow.com/questions/7522250/
    include_package_data=True,
    # This is a trick to avoid duplicating dependencies between both setup.py and
    # requirements.txt.
    # requirements.txt must be included in MANIFEST.in for this to work.
    # It does not work for all types of dependencies (e.g. VCS dependencies).
    # For VCS dependencies, use pip >= 19 and the PEP 508 syntax.
    #   Example: 'requests @ git+https://github.com/requests/requests.git@branch_or_tag'
    #   See: https://github.com/pypa/pip/issues/6162
    install_requires=project_dir.joinpath("requirements.txt").read_text().split("\n"),
    zip_safe=True,
    license="MIT",
    license_files=["LICENSE.txt"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
    ],
    entry_points={"console_scripts": ["PySimultan=PySimultan.cli:main"]},
)



from .ParameterStructure.Components import SimSlot, SimSlotBase, ComponentUtils
# from .ParameterStructure.Components.ComponentUtils import *

"""
Valid slots

COMP_SLOT_ABGABE = "Abgabe"
COMP_SLOT_AREAS = "Geometrische_Flächen"
COMP_SLOT_CALCULATION = "Berechnung"
COMP_SLOT_COMMUNICATION = "Kommunikation"
COMP_SLOT_COMPOSITE = "Aufbau"
COMP_SLOT_CONNECTED_TO = "Angeschlossen_an"
COMP_SLOT_COST = "Kosten"
COMP_SLOT_DELIMITER = "_0"
COMP_SLOT_ERZEUGER = "Erzeuger"
COMP_SLOT_IMPORT = "Import"
COMP_SLOT_ITEM = "Element"
COMP_SLOT_JOINT = "Anschluss"
COMP_SLOT_LAYER = "Schicht"
COMP_SLOT_LENGTHS = "Geometrische_Längen"
COMP_SLOT_LIST = "Liste"
COMP_SLOT_MATERIAL = "Material"
COMP_SLOT_OBJECT = "Geometrisches_Objekt"
COMP_SLOT_OPENING = "Öffnung"
COMP_SLOT_POSITION = "Verortung"
COMP_SLOT_REGULATION = "Anforderungen"
COMP_SLOT_SINGLE_ACOUSTICS = "Akustik"
COMP_SLOT_SINGLE_COOLING = "Kühlung"
COMP_SLOT_SINGLE_ELECTRICAL = "Elektro"
COMP_SLOT_SINGLE_FIRE_SAFETY = "Brandschutz"
COMP_SLOT_SINGLE_HEATIG = "Heizung"
COMP_SLOT_SINGLE_HUMIDITY = "Feuchte"
COMP_SLOT_SINGLE_LIGHT_ARTIF = "Kunstlicht"
COMP_SLOT_SINGLE_LIGHT_NATURAL = "Naturlicht"
COMP_SLOT_SINGLE_MSR = "MSR"
COMP_SLOT_SINGLE_WASTE = "Abwasser"
COMP_SLOT_SINGLE_WATER = "Wasser"
COMP_SLOT_SIZE = "Geometrische_Maße"
COMP_SLOT_SPECIFICATION = "Leistungsbeschr"
COMP_SLOT_SYSTEM = "System"
COMP_SLOT_TUPLE = "Tupel"
COMP_SLOT_UNDEFINED = "Undefined Slot"
COMP_SLOT_VERTEILER = "Verteiler"
COMP_SLOT_VERTEILER_PART = "Verteiler_Teil"
COMP_SLOT_VERTEILER_PIPE = "Verteiler_Kanal"
COMP_SLOT_VOLUMES = "Geometrische_Volumina
"""


def create_slot(slot_name, slot_extension):
    return SimSlot(SimSlotBase(slot_name), str(slot_extension))
